"""
cfn-lsp-extra update-specification
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/cfn-resource-specification-format.html
https://github.com/emacs-lsp/lsp-mode/issues/3676
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-properties-amplifyuibuilder-component-predicate.html

Not in the spec are attribute descriptions, and allowed values.
"""

from __future__ import annotations

import functools
import json  # noqa: I001
import logging
import os
import tempfile
from pathlib import Path
from typing import Callable, Optional

import requests
from bs4 import BeautifulSoup
from markdownify import markdownify as md  # type: ignore[import-untyped]

from .. import remove_prefix, remove_suffix
from ..aws_data import AWSSpecification, Tree
from .with_success_failure_count import WithSuccessFailureCount

ALLOWED_VALUES_PREFIX = "*Allowed values*:"
MAX_ALLOWED_VALUES_WIDTH = 30
DEFAULT_SPEC_URL = "https://dnwj8swjjbsbt.cloudfront.net/latest/gzip/CloudFormationResourceSpecification.json"
PARSE_SUCCESS_RATIO_KEY = "parse_success_ratio"

logger = logging.getLogger(__name__)


def md_(s: str) -> str:
    return md(s).replace("\n\n", "\n")  # type: ignore[no-any-return]


def to_aws_context(
    d: Tree, parent: Optional[str], base_directory: Path
) -> WithSuccessFailureCount[Tree]:
    if not isinstance(d, dict):
        return WithSuccessFailureCount.zero(d)
    d_: WithSuccessFailureCount[Tree] = WithSuccessFailureCount.zero({})
    for k, v in d.items():
        if k == AWSSpecification.DOCUMENTATION:
            content = file_content(base_directory, v)
            doc_with_counts = documentation(content, v, parent)
            doc = d_.value[
                AWSSpecification.MARKDOWN_DOCUMENTATION
            ] = doc_with_counts.value
            d_ = d_.add_counts(doc_with_counts)
            d_.value[AWSSpecification.REF_RETURN_VALUE] = ref_return_value(
                content, v.split("/")[-1]
            )
            lines = doc.splitlines()
            idx, allowed_values_line = next(
                (
                    (idx, line)
                    for idx, line in enumerate(lines)
                    if line.startswith(ALLOWED_VALUES_PREFIX)
                ),
                (None, None),
            )
            if allowed_values_line and idx:
                allowed_values = (
                    remove_prefix(allowed_values_line, ALLOWED_VALUES_PREFIX)
                    .strip("` ")
                    .replace(" | ", "|")
                    .split("|")
                )
                d_.value[AWSSpecification.ALLOWED_VALUES] = allowed_values
                if len(allowed_values_line) > MAX_ALLOWED_VALUES_WIDTH:
                    new_line = (
                        " | ".join(allowed_values)[: MAX_ALLOWED_VALUES_WIDTH - 3]
                        + "..."
                    )
                    lines[idx] = f"{ALLOWED_VALUES_PREFIX} `{new_line}`"
                    d_.value[AWSSpecification.MARKDOWN_DOCUMENTATION] = "\n".join(lines)
        if (
            isinstance(v, dict)
            and AWSSpecification.ATTRIBUTES in v
            and AWSSpecification.DOCUMENTATION in v
        ):
            set_attribute_doc(
                v[AWSSpecification.DOCUMENTATION],
                v[AWSSpecification.ATTRIBUTES],
            )
        sub_d_with_counts = to_aws_context(v, k, base_directory)
        d_.value[k] = sub_d_with_counts.value
        d_ = d_.add_counts(sub_d_with_counts)
    return d_


def set_attribute_doc(base_link: str, attribs_dct: Tree) -> None:
    for attrib, attrib_dct in attribs_dct.items():
        id_ = f"{base_link}#{attrib}-fn::getatt"
        attrib_dct[AWSSpecification.DOCUMENTATION] = id_


@functools.lru_cache(maxsize=None)
def file_content(
    base_directory: Path,
    link: str,
    on_fail_try_download: bool = True,
) -> BeautifulSoup:
    def read_file(loc: Path) -> BeautifulSoup:
        return BeautifulSoup(loc.read_text(), features="lxml")

    location = link.split("/")[-1].lstrip("/")
    if "#" in link:  # Subprop
        sub_path, _, id_ = location.partition("#")
        file_path = base_directory / sub_path
        try:
            return read_file(file_path)
        except FileNotFoundError:
            if on_fail_try_download:
                try_download(link, file_path)
                return file_content(base_directory, link, on_fail_try_download=False)
            logger.info("No file found for %s", link)
            return BeautifulSoup("")
    else:  # resource
        file_path = base_directory / location
        logger.debug("Processing %s", file_path)
        try:
            return read_file(file_path)
        except FileNotFoundError:
            if on_fail_try_download:
                try_download(link, file_path)
                return file_content(base_directory, link, on_fail_try_download=False)
            logger.info("No file found for %s", link)
            return BeautifulSoup("")


def documentation(
    content: BeautifulSoup, link: str, parent: Optional[str]
) -> WithSuccessFailureCount[str]:
    if "#" in link:  # Subprop
        # Sometimes the link is out of date (redirect), we guess the true id link using the h1 attrib
        split = link.split("#")
        id_, url_split = split[-1], split[0].split("/")
        replace: Callable[[str], str] = lambda s: s.replace(
            "aws-properties", "cfn"
        ).replace("aws-resource", "cfn")
        old_id_prefix = replace(remove_suffix(url_split[-1], ".html"))
        header = content.find("h1")
        if header and hasattr(header, "attrs"):
            true_id_prefix = replace(header.attrs["id"])
            true_id = id_.replace(old_id_prefix, true_id_prefix)
            if true_id != id_:
                logger.info("Replaced %s with %s", id_, true_id)
        else:
            true_id = id_
        dt = content.find("dt", {"id": true_id})
        if not dt:
            logger.error(
                "No documentation found for prop %s using id %s", link, true_id
            )
            return WithSuccessFailureCount.failure("")
        dd = dt.findNext("dd")
        doc = md_(str(dd)) if dd else ""
    else:  # resource
        div = content.find("div", {"class": "awsdocs-page-header-container"})
        tags, el = [], div
        # Add everything until the next header tag
        while True:
            if not el:
                break
            el = el.find_next_sibling()
            if el and el.name and not (el.name.startswith("h") and len(el.name) == 2):
                tags.append(el)
            else:
                break
        doc = md_("".join(map(str, tags))) if tags else ""

    s = f"`{parent if parent else ''}`\n{doc}"
    if doc:
        return WithSuccessFailureCount.success(s)
    logger.error("Documentation for %s was found to be empty", link)
    return WithSuccessFailureCount.failure(s)


def ref_return_value(content: BeautifulSoup, slug: str) -> str:
    prefix = remove_suffix(slug, ".html")
    id_ = f"{prefix}-return-values-ref"
    div = content.find("h3", {"id": id_})
    p_tags, el = [], div
    while True:
        if not el:
            break
        el = el.find_next_sibling()
        if el and el.name and el.name == "p":
            p_tags.append(el)
        else:
            break
    return md_("".join(map(str, p_tags))) if p_tags else ""


def run_command(cmd: str) -> None:
    logger.info("Running '%s'", cmd)
    os.system(cmd)


def try_download(url: str, out_file_name: Path) -> None:
    run_command(f"curl -L -X GET {url} > {out_file_name.absolute()}")


def run(
    spec_file: Optional[Path],
    documentation_directory: Optional[Path],
    out_path: Optional[Path],
) -> None:
    out_path = out_path or Path("new-aws-context.json").absolute()
    if spec_file:
        spec_json = json.loads(spec_file.read_bytes())
    else:
        logger.info("Downloading spec from %s", DEFAULT_SPEC_URL)
        # requests handles gzip OOTB unlike urllib
        spec_json = json.loads(requests.get(DEFAULT_SPEC_URL).text)
    with tempfile.TemporaryDirectory() as tmp_directory:
        os.chdir(tmp_directory)
        doc_dir = documentation_directory or (
            Path("docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide")
        )
        if not documentation_directory:
            logger.info("Downloading documentation from %s...", doc_dir)
            run_command(f"wget --no-parent -r https://{doc_dir}")
        else:
            logger.info(
                "Not downloading documentation, using existing directory %s",
                documentation_directory,
            )
        ctx_map, md_succ, md_fails = to_aws_context(spec_json, None, doc_dir).to_tuple()
        ctx_map[PARSE_SUCCESS_RATIO_KEY] = round(md_succ / (md_fails + md_succ), 4)
        logger.info("Failed getting markdown: %d/%d", md_fails, md_fails + md_succ)
        with open(out_path, "w") as f_:
            json.dump(ctx_map, f_, indent=2)
    logger.info("Wrote context to %s", out_path)



"""
Completions for !Refs
"""
from typing import Optional

from lsprotocol.types import CompletionItem, CompletionList, Position
from pygls.workspace import Document

from ..aws_data import AWSContext, AWSRefName, AWSRefSource, Tree
from ..cursor import text_edit, word_before_after_position
from ..decode.extractors import Extractor
from ..ref import REF_EXTRACTOR, REF_SRC_EXTRACTOR


def ref_completions(
    template_data: Tree,
    document: Document,
    position: Position,
    aws_context: AWSContext,
    ref_extractor: Extractor[AWSRefName] = REF_EXTRACTOR,
    ref_src_extractor: Extractor[AWSRefSource] = REF_SRC_EXTRACTOR,
) -> Optional[CompletionList]:
    ref_lookup = ref_extractor.extract(template_data)
    ref_span = ref_lookup.at(position.line, position.character)
    if ref_span:
        before, after = word_before_after_position(document, position)
        ref_src_lookup = ref_src_extractor.extract(template_data)
        items = [
            CompletionItem(
                label=ref_src.logical_name,
                documentation=ref_src.as_documentation(aws_context),
                text_edit=text_edit(position, before, after, ref_src.logical_name),
            )
            for ref_src, _ in ref_src_lookup.items()
        ]
        return CompletionList(is_incomplete=False, items=items)
    return None



import re
from typing import Pattern, Tuple

from lsprotocol.types import Position, Range, TextEdit
from pygls.workspace import Document

RE_END_WORD = re.compile("^[A-Za-z_0-9!:]*")
RE_START_WORD = re.compile("[A-Za-z_0-9!:]*$")

def text_edit(position: Position, before: str, after: str, text: str) -> TextEdit:
    """Return a TextEdit for position given text before and after the cursor."""
    start = Position(line=position.line, character=position.character - len(before))
    end = Position(line=position.line, character=position.character + len(after))
    return TextEdit(range=Range(start=start, end=end), new_text=text)


def word_before_after_position(
    document: Document,
    position: Position,
    re_start_word: Pattern[str] = RE_START_WORD,
    re_end_word: Pattern[str] = RE_END_WORD,
) -> Tuple[str, str]:
    lines = document.lines
    if position.line >= len(lines):
        return "", ""

    pos = document.position_codec.position_from_client_units(lines, position)
    row, col = pos.line, pos.character

    line = lines[row]
    # Split word in two
    start = line[:col]
    end = line[col:]

    # Take end of start and start of end to find word
    # These are guaranteed to match, even if they match the empty string
    m_start = re_start_word.findall(start)
    m_end = re_end_word.findall(end)

    return m_start[0], m_end[-1]


def word_at_position_char_bounds(
    document: Document,
    position: Position,
    re_start_word: Pattern[str] = RE_START_WORD,
    re_end_word: Pattern[str] = RE_END_WORD,
) -> Tuple[int, int]:
    before, after = word_before_after_position(
        document, position, re_start_word, re_end_word
    )
    char = position.character
    return char - len(before), char + len(after)



"""
Classes for extracting positional information from decoded documents.

For example extracting the positions of resource parameters.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Callable, Generic, List, Optional, Set, Tuple, TypeVar, Union

from attrs import frozen

from .. import remove_prefix
from ..aws_data import (
    AWSLogicalId,
    AWSParameter,
    AWSPropertyName,
    AWSResourceName,
    Tree,
)
from . import DEBUG_CHAR
from .position import PositionLookup, Spanning
from .yaml_decoding import POSITION_PREFIX, VALUES_POSITION_PREFIX

E = TypeVar("E", covariant=True)


class Extractor(ABC, Generic[E]):
    @abstractmethod
    def extract(self, node: Tree) -> PositionLookup[E]:
        """Call extract contents from node.

        Parameters
        ----------
        node : Tree
            The root node to extract from.

        Returns
        -------
        PositionLookup[T]
            A PositionLookup object containing items from source."""
        ...


class RecursiveExtractor(Extractor[E]):
    def extract(self, node: Tree) -> PositionLookup[E]:
        """Call extract_node at each of the inner nodes of node.

        Parameters
        ----------
        node : Tree
            node to extract from (recursively).

        Returns
        -------
        PositionLookup[T]
            A PositionLookup object containing items from source."""
        if isinstance(node, dict):
            position_lookup = PositionLookup.from_iterable(self.extract_node(node))
            iterable = node.values()
        else:
            position_lookup = PositionLookup[E]()
            iterable = node
        for child in iterable:
            if isinstance(child, dict):
                position_lookup.extend_with_appends(self.extract(child))
            elif isinstance(child, list):
                for sub_child in filter(lambda c: isinstance(c, (dict, list)), child):
                    position_lookup.extend_with_appends(self.extract(sub_child))
        return position_lookup

    @abstractmethod
    def extract_node(self, node: Tree) -> List[Spanning[E]]:
        ...


class ResourcePropertyExtractor(Extractor[AWSPropertyName]):
    """Extractor for resource and nested properties.

    Methods
    -------
    extract(node)
        Extract resource and nested properties from node."""

    def extract(self, node: Tree) -> PositionLookup[AWSPropertyName]:
        props = []
        if "Resources" in node and isinstance(node["Resources"], dict):
            for resource in filter(
                lambda r: isinstance(r, dict), node["Resources"].values()
            ):
                is_res_node = "Properties" in resource and "Type" in resource
                if is_res_node and isinstance(resource["Properties"], dict):
                    type_ = resource["Type"] or ""
                    props.extend(
                        self._extract_recursive(
                            resource["Properties"],
                            AWSResourceName(value=type_),
                        )
                    )
                elif (
                    is_res_node
                    and isinstance(resource["Properties"], str)
                    and VALUES_POSITION_PREFIX in resource
                ):
                    props.extend(
                        self._extract_unfinished(
                            resource,
                            resource["Properties"],
                            AWSResourceName(value=resource.get("Type", "")),
                        )
                    )
        return PositionLookup.from_iterable(props)

    def _extract_recursive(
        self, node: Tree, parent: Union[AWSPropertyName, AWSResourceName]
    ) -> List[Spanning[AWSPropertyName]]:
        props = []
        if isinstance(node, list):
            for sub_node in node:
                if isinstance(sub_node, (dict, list)):
                    props.extend(self._extract_recursive(sub_node, parent))
            return props
        for key, value in node.items():
            if key.startswith(POSITION_PREFIX):
                prop = remove_prefix(key, POSITION_PREFIX)
                aws_prop = parent / prop
                line, char = value
                props.append(
                    Spanning[AWSPropertyName](
                        value=aws_prop, line=line, char=char, span=len(prop)
                    )
                )

                if isinstance(node[prop], dict):
                    props.extend(self._extract_recursive(node[prop], aws_prop))
                elif isinstance(node[prop], list):
                    for sub_node in filter(lambda p: isinstance(p, dict), node[prop]):
                        if isinstance(sub_node, dict):
                            props.extend(self._extract_recursive(sub_node, aws_prop))
                        elif isinstance(sub_node, list):
                            props.extend(self._extract_recursive(sub_node, parent))
                elif node[prop] == DEBUG_CHAR and VALUES_POSITION_PREFIX in node:
                    props.extend(
                        self._extract_unfinished(node, node[prop], parent / prop)
                    )
        return props

    def _extract_unfinished(
        self,
        node: Tree,
        unfinished_property: Tree,
        parent: Union[AWSPropertyName, AWSResourceName],
    ) -> List[Spanning[AWSPropertyName]]:
        props = []
        for dct in node[VALUES_POSITION_PREFIX]:
            key = POSITION_PREFIX + unfinished_property
            if key in dct:
                line, char = dct[key]
                aws_prop = parent / unfinished_property
                props.append(
                    Spanning[AWSPropertyName](
                        value=aws_prop,
                        line=line,
                        char=char,
                        span=len(unfinished_property),
                    )
                )
                break
        return props


class ResourceExtractor(Extractor[AWSResourceName]):
    """Extractor for resources names.

    Methods
    -------
    extract(node)
        Extract resource names from node."""

    def extract(self, node: Tree) -> PositionLookup[AWSResourceName]:
        props = []
        if "Resources" in node and isinstance(node["Resources"], dict):
            for resource_dct in node["Resources"].values():
                if (
                    isinstance(resource_dct, dict)
                    and "Type" in resource_dct
                    and VALUES_POSITION_PREFIX in resource_dct
                ):
                    type_ = resource_dct["Type"] or ""
                    value_positions = resource_dct[VALUES_POSITION_PREFIX]
                    key = POSITION_PREFIX + str(type_)  # type could be fn call
                    for dct in value_positions:
                        if key in dct:
                            line, char = dct[key]
                            props.append(
                                Spanning[AWSResourceName](
                                    value=AWSResourceName(value=type_),
                                    line=line,
                                    char=char,
                                    span=len(type_),
                                )
                            )
                            break
        return PositionLookup.from_iterable(props)


@frozen
class StaticPath:
    value: Tuple[str, ...]

    MatchAny: str = "*"

    def __truediv__(self, sub_node: str) -> StaticPath:
        return StaticPath(value=self.value + (sub_node,))

    def head_tail(self) -> Tuple[Optional[str], StaticPath]:
        if not self.value:
            return None, StaticPath(value=())
        head, *tail = self.value
        return head, StaticPath(value=tuple(tail))

    @staticmethod
    def root(key: str) -> StaticPath:
        return StaticPath(value=(key,))


class StaticExtractor(Extractor[StaticPath]):
    """Extractor matching a set of 'fixed' paths.

    Methods
    -------
    extract(node)
        Extract fixed paths node."""

    def __init__(self, paths: Set[StaticPath]):
        self.paths = paths

    def extract(self, node: Tree) -> PositionLookup[StaticPath]:
        spans = (
            span
            for path in self.paths
            for span in self._extract_recursive(node, path, path)
        )
        return PositionLookup.from_iterable(spans)

    def _extract_recursive(
        self, node: Tree, cur_path: StaticPath, full_path: StaticPath
    ) -> List[Spanning[StaticPath]]:
        spans = []
        head, tail = cur_path.head_tail()
        is_base_case, match_all = not tail.value, head == StaticPath.MatchAny
        if is_base_case and match_all:
            for key in node:
                pos_key = POSITION_PREFIX + key
                if pos_key in node:
                    line, char = node[pos_key]
                    spans.append(
                        Spanning[StaticPath](
                            value=full_path, line=line, char=char, span=len(key)
                        )
                    )
        elif is_base_case and head:
            pos_key = POSITION_PREFIX + head
            if pos_key in node:
                line, char = node[pos_key]
                spans.append(
                    Spanning[StaticPath](
                        value=full_path, line=line, char=char, span=len(head)
                    )
                )
        elif match_all:
            for sub_node in node.values():
                if isinstance(sub_node, dict):
                    spans.extend(self._extract_recursive(sub_node, tail, full_path))
        elif head and head in node and isinstance(node[head], dict):
            spans.extend(self._extract_recursive(node[head], tail, full_path))
        return spans


class AllowedValuesExtractor(Extractor[AWSPropertyName]):
    """Extractor for property values.

    Methods
    -------
    extract(node)
        Extract resource and nested property values from node."""

    def extract(self, node: Tree) -> PositionLookup[AWSPropertyName]:
        props = []
        if "Resources" in node and isinstance(node["Resources"], dict):
            for resource in filter(
                lambda r: isinstance(r, dict), node["Resources"].values()
            ):
                is_res_node = "Properties" in resource and "Type" in resource
                if is_res_node and isinstance(resource["Properties"], dict):
                    type_ = resource["Type"] or ""
                    props.extend(
                        self._extract_recursive(
                            resource["Properties"],
                            AWSResourceName(value=type_),
                        )
                    )
                # don't care if isinstance(resource["Properties"], str) since this means no values
        return PositionLookup.from_iterable(props)

    def _extract_recursive(
        self, node: Tree, parent: Union[AWSPropertyName, AWSResourceName]
    ) -> List[Spanning[AWSPropertyName]]:
        props = []
        if isinstance(node, list):
            for sub_node in node:
                if isinstance(sub_node, (dict, list)):
                    props.extend(self._extract_recursive(sub_node, parent))
            return props

        values_positions = node.get(VALUES_POSITION_PREFIX, [])
        for key in node:
            prop = remove_prefix(key, POSITION_PREFIX)
            aws_prop = parent / prop
            if key.startswith(POSITION_PREFIX):
                # Try to obtain value position from values_positions
                value = str(node.get(prop, ""))
                val_key = POSITION_PREFIX + value
                for dct in values_positions:
                    if val_key in dct:
                        line, char = dct[val_key]
                        props.append(
                            Spanning[AWSPropertyName](
                                value=aws_prop,
                                line=line,
                                char=char,
                                span=len(value),
                            )
                        )
                        break

            if prop in node and isinstance(node[prop], dict):
                props.extend(self._extract_recursive(node[prop], aws_prop))
            elif prop in node and isinstance(node[prop], list):
                for sub_node in filter(lambda p: isinstance(p, dict), node[prop]):
                    if isinstance(sub_node, dict):
                        props.extend(self._extract_recursive(sub_node, aws_prop))
                    elif isinstance(sub_node, list):
                        props.extend(self._extract_recursive(sub_node, parent))
        return props


class ParameterExtractor(Extractor[AWSParameter]):
    """Extractor for parameters.

    Methods
    -------
    extract(node)
        Extract resource names from node."""

    SECTION = "Parameters"

    def extract(self, node: Tree) -> PositionLookup[AWSParameter]:
        params = []
        if self.SECTION in node and isinstance(node[self.SECTION], dict):
            for param_name, content_dct in node[self.SECTION].items():
                key = POSITION_PREFIX + param_name
                add_parameter = (
                    key in node[self.SECTION]
                    and isinstance(content_dct, dict)
                    and "Type" in content_dct
                )
                if add_parameter:
                    type_ = content_dct["Type"]
                    line, char = node[self.SECTION][key]
                    param = AWSParameter(
                        logical_name=param_name,
                        type_=type_,
                        description=content_dct.get("Description", None),
                        default=content_dct.get("Default", None),
                    )
                    params.append(
                        Spanning[AWSParameter](
                            value=param,
                            line=line,
                            char=char,
                            span=len(param_name),
                        )
                    )
        return PositionLookup.from_iterable(params)


K = TypeVar("K", covariant=True)


class KeySetExtractor(RecursiveExtractor[K]):
    """Extractor for named key values.

    Methods
    -------
    extract(node)
        Extract keys from a given set from node."""

    def __init__(self, key_names: Set[str], name_fn: Callable[[str], K]):
        self.key_names = key_names
        self.name_fn = name_fn

    def extract_node(self, node: Tree) -> List[Spanning[K]]:
        found = []
        for key, value in node.items():
            if key in self.key_names and VALUES_POSITION_PREFIX in node:
                if key == "Fn::GetAtt" and isinstance(value, list):
                    value = ".".join(map(str, value))
                for val_pos_dct in node[VALUES_POSITION_PREFIX]:
                    p_key = POSITION_PREFIX + str(value)
                    if p_key in val_pos_dct:
                        line, char = val_pos_dct[p_key]
                        found.append(
                            Spanning[K](
                                value=self.name_fn(value),
                                line=line,
                                char=char,
                                span=len(value),
                            )
                        )
        return found


class KeyExtractor(KeySetExtractor[K]):
    """Extractor for named key values.

    Methods
    -------
    extract(node)
        Extract keys from node."""

    def __init__(self, key_name: str, name_fn: Callable[[str], K]):
        super().__init__({key_name}, name_fn)


class GetAttExtractor(RecursiveExtractor[str]):
    """Extractor for GetAtts.

    Methods
    -------
    extract(node)
        Extract GetAtts from node."""

    KEY = "Fn::GetAtt"

    def extract_node(self, node: Tree) -> List[Spanning[str]]:
        found = []
        for key, value in node.items():
            if key == self.KEY and VALUES_POSITION_PREFIX in node:
                expr = ".".join(value)
                for val_pos_dct in node[VALUES_POSITION_PREFIX]:
                    p_key = POSITION_PREFIX + expr
                    if p_key in val_pos_dct:
                        line, char = val_pos_dct[p_key]
                        found.append(
                            Spanning[str](
                                value=expr,
                                line=line,
                                char=char,
                                span=len(expr),
                            )
                        )
        return found


class LogicalIdExtractor(Extractor[AWSLogicalId]):
    """Extractor for the logical ids of a template.

    Methods
    -------
    extract(node)
        Extract logical ids from node."""

    SECTION = "Resources"

    def extract(self, node: Tree) -> PositionLookup[AWSLogicalId]:
        params = []
        if self.SECTION in node and isinstance(node[self.SECTION], dict):
            for logical_id, content_dct in node[self.SECTION].items():
                key = POSITION_PREFIX + logical_id
                if key in node[self.SECTION]:
                    if isinstance(content_dct, dict) and "Type" in content_dct:
                        type_ = content_dct.get("Type", None)
                    else:
                        type_ = None
                    line, char = node[self.SECTION][key]
                    params.append(
                        Spanning[AWSLogicalId](
                            value=AWSLogicalId(logical_name=logical_id, type_=type_),
                            line=line,
                            char=char,
                            span=len(logical_id),
                        )
                    )

        return PositionLookup.from_iterable(params)


T = TypeVar("T", covariant=True)


class CompositeExtractor(Extractor[T]):
    """Accumulates the results of multiple Extractor objects."""

    def __init__(self, *extractors: Extractor[T]):
        self._extractors = extractors

    def extract(self, node: Tree) -> PositionLookup[T]:
        lookup = PositionLookup[T]()
        for extractor in self._extractors:
            lookup.extend_with_appends(extractor.extract(node))
        return lookup



from __future__ import annotations

from typing import Dict, Generic, Iterable, List, Optional, Tuple, TypeVar

from attrs import frozen

E = TypeVar("E")


@frozen
class Spanning(Generic[E]):
    """The enrichment of some object with a location span in a document.

    Attributes
    ----------
    value : E
        The value this span wraps
    line : int
        Line position marking the start of the span.
    char : int
        Character position marking the start of the span.
    span : int
        The length of the span"""

    value: E
    line: int
    char: int
    span: int


T = TypeVar("T")
PositionList = List[Tuple[int, int, int]]


class PositionLookup(Dict[T, PositionList]):
    """A thin wrapper around Dict[T, List[Tuple[int, int, int]]]."""

    def __missing__(self, key: T) -> PositionList:
        self[key] = []
        return self[key]

    def at(self, line: int, char: int) -> Optional[Spanning[T]]:
        for item, positions in self.items():
            for item_line, char_min, item_span in positions:
                char_max = char_min + item_span
                within_col = char_min <= char <= char_max
                if line == item_line and within_col:
                    return Spanning[T](
                        value=item, line=item_line, char=char_min, span=item_span
                    )
        return None

    def extend_with_appends(self, other: PositionLookup[T]) -> None:
        for key, value in other.items():
            self[key].extend(value)

    @classmethod
    def from_iterable(cls, iterable: Iterable[Spanning[T]]) -> PositionLookup[T]:
        lookup: PositionLookup[T] = cls()
        for span in iterable:
            lookup[span.value].append((span.line, span.char, span.span))
        return lookup


ST = TypeVar("ST")
TT = TypeVar("TT")


@frozen
class PositionLink(Generic[ST, TT]):
    """The linking of a source object and a target object in a document.

    Attributes
    ----------
    source_span : Spanning[ST]
        The source object
    target_span : Spanning[TT]
        The target object"""

    source_span: Spanning[ST]
    target_span: Spanning[TT]



# type: ignore
# flake8: noqa
"""
Utilities for parsing json document strings.
"""
import bisect
import json
from json import JSONDecodeError
from json import scanner
from json.decoder import WHITESPACE
from json.decoder import WHITESPACE_STR
from json.decoder import JSONDecoder
from json.decoder import scanstring
from json.scanner import py_make_scanner
from typing import Any
from typing import List

from .yaml_decoding import POSITION_PREFIX
from .yaml_decoding import VALUES_POSITION_PREFIX


# This is a slightly modified version of json.decoder.JSONObject
# added lines have comments starting with 'EDIT:'
def cfn_json_object(
    s_and_end,
    strict,
    scan_once,
    object_hook,
    object_pairs_hook,
    memo=None,
    _w=WHITESPACE.match,
    _ws=WHITESPACE_STR,
):  # pragma: no cover
    s, end = s_and_end
    init = end
    pairs = []
    pairs_append = pairs.append
    # Backwards compatibility
    if memo is None:
        memo = {}
    memo_get = memo.setdefault
    # Use a slice to prevent IndexError from being raised, the following
    # check will raise a more specific ValueError if the string is empty
    nextchar = s[end : end + 1]
    # Normally we expect nextchar == '"'
    if nextchar != '"':
        if nextchar in _ws:
            end = _w(s, end).end()
            nextchar = s[end : end + 1]
        # Trivial empty object
        if nextchar == "}":
            if object_pairs_hook is not None:
                result = object_pairs_hook(pairs)
                return result, end + 1
            pairs = {}
            if object_hook is not None:
                pairs = object_hook(pairs)
            return pairs, end + 1
        elif nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end
            )
    end += 1
    while True:
        key_start = end  # EDIT: gets start position of key
        key, end = scanstring(s, end, strict)
        key_end = end  # EDIT: gets end position of key
        key = memo_get(key, key)
        # To skip some function call overhead we optimize the fast paths where
        # the JSON key separator is ": " or just ":".
        if s[end : end + 1] != ":":
            end = _w(s, end).end()
            if s[end : end + 1] != ":":
                raise JSONDecodeError("Expecting ':' delimiter", s, end)
        end += 1

        try:
            if s[end] in _ws:
                end += 1
                if s[end] in _ws:
                    end = _w(s, end + 1).end()
        except IndexError:
            pass

        try:
            value_start = end
            value, end = scan_once(s, end)
        except StopIteration as err:
            raise JSONDecodeError("Expecting value", s, err.value) from None
        pairs_append((key, value))
        # EDIT: adds positional data to result
        if isinstance(value, str):
            if VALUES_POSITION_PREFIX not in {k for k, _ in pairs}:
                values_lst = []
                pairs_append((VALUES_POSITION_PREFIX, values_lst))
            else:
                values_lst = next(v for k, v in pairs if k == VALUES_POSITION_PREFIX)
            values_lst.append({POSITION_PREFIX + value: [value_start + 1, end + 1]})
        # END EDIT
        pairs_append((POSITION_PREFIX + key, [key_start, key_end]))
        try:
            nextchar = s[end]
            if nextchar in _ws:
                end = _w(s, end + 1).end()
                nextchar = s[end]
        except IndexError:
            nextchar = ""
        end += 1

        if nextchar == "}":
            break
        elif nextchar != ",":
            raise JSONDecodeError("Expecting ',' delimiter", s, end - 1)
        end = _w(s, end).end()
        nextchar = s[end : end + 1]
        end += 1
        if nextchar != '"':
            raise JSONDecodeError(
                "Expecting property name enclosed in double quotes", s, end - 1
            )
    if object_pairs_hook is not None:
        result = object_pairs_hook(pairs)
        return result, end
    pairs = dict(pairs)
    if object_hook is not None:
        pairs = object_hook(pairs)
    return pairs, end


class CfnJSONDecoder(JSONDecoder):
    """A json decoder which saves positional information of elements."""

    def __init__(self):
        super().__init__()
        self.parse_object = cfn_json_object
        self.scan_once = py_make_scanner(self)

    def raw_decode(self, s: str, idx: int = 0):
        dct, end = super().raw_decode(s, idx)
        line_starts = [0]
        cnt = 0
        for line in s.splitlines():
            cnt += len(line) + 1  # + 1 for the newline character
            line_starts.append(cnt)
        self._convert_abs_char_to_line_char(dct, line_starts)
        return dct, end

    def _convert_abs_char_to_line_char(self, d: Any, line_starts: List[int]):
        for key, value in d.items():
            if key.startswith(POSITION_PREFIX):
                c_start, _ = value
                line = bisect.bisect_left(line_starts, c_start)
                char = (
                    0
                    if c_start == line_starts[line]
                    else c_start - line_starts[line - 1]
                )
                d[key] = [line - 1, char]
            elif isinstance(value, dict):
                self._convert_abs_char_to_line_char(value, line_starts)
            elif isinstance(value, list):
                for e in value:
                    if isinstance(e, dict):
                        self._convert_abs_char_to_line_char(e, line_starts)



"""
Definitions for !GettAtts.
"""

from typing import Optional

from lsprotocol.types import Location, Position, Range
from pygls.workspace import Document

from ..aws_data import AWSContext, AWSRefName, Tree
from ..decode.extractors import KeyExtractor, LogicalIdExtractor
from ..ref import resolve_ref

ATTRIBUTE_EXTRACTOR = KeyExtractor[AWSRefName](
    "Fn::GetAtt", lambda s: AWSRefName(value=s.split(".")[0])
)
ATTRIBUTE_SRC_EXTRACTOR = LogicalIdExtractor()


def attribute_definition(
    template_data: Tree,
    document: Document,
    position: Position,
    aws_context: AWSContext,
) -> Optional[Location]:
    link = resolve_ref(
        position,
        template_data,
        ref_extractor=ATTRIBUTE_EXTRACTOR,
        ref_src_extractor=ATTRIBUTE_SRC_EXTRACTOR,
    )
    if link:
        return Location(
            uri=document.uri,
            range=Range(
                start=Position(
                    line=link.source_span.line, character=link.source_span.char
                ),
                end=Position(
                    line=link.source_span.line,
                    character=link.source_span.char + link.source_span.span,
                ),
            ),
        )
    return None



"""
Intrinsic functions, see:
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html
For more information.
"""

from attrs import frozen


@frozen
class IntrinsicFunction:
    function: str
    documentation: str
    full_name_prefix: str = "Fn::"
    short_form_prefix: str = "!"

    def short_form(self) -> str:
        return self.short_form_prefix + self.function

    def full_name(self) -> str:
        return self.full_name_prefix + self.function


# Hard coded temporarily (?)
INTRINSIC_FUNCTIONS = [
    IntrinsicFunction(
        function="Base64",
        documentation="The intrinsic function `Fn::Base64` returns the Base64 representation of the  \ninput string\\. This function is typically used to pass encoded data to Amazon  \nEC2 instances by way of the `UserData` property\\.",
    ),
    IntrinsicFunction(
        function="Cidr",
        documentation="The intrinsic function `Fn::Cidr` returns an array of CIDR address blocks\\. The  \nnumber of CIDR blocks returned is dependent on the `count` parameter\\.",
    ),
    IntrinsicFunction(
        function="FindInMap",
        documentation="The intrinsic function `Fn::FindInMap` returns the value corresponding to keys  \nin a two\\-level map that's declared in the `Mappings` section\\.",
    ),
    IntrinsicFunction(
        function="GetAtt",
        documentation="The `Fn::GetAtt` intrinsic function returns the value of an attribute from a  \nresource in the template\\. When the [`AWS::LanguageExtensions` transform](transform-aws-languageextensions.md)  \ntransform is used, you can use intrinsic functions as a parameter to  \n`Fn::GetAtt`\\. For more information about `GetAtt` return values for a  \nparticular resource, refer to the documentation for that resource in the  \n[Resource and property reference](aws-template-resource-type-ref.md)\\.",
    ),
    IntrinsicFunction(
        function="GetAZs",
        documentation="The intrinsic function `Fn::GetAZs` returns an array that lists Availability  \nZones for a specified region in alphabetical order\\. Because customers have  \naccess to different Availability Zones, the intrinsic function `Fn::GetAZs`  \nenables template authors to write templates that adapt to the calling user's  \naccess\\. That way you don't have to hard\\-code a full list of Availability Zones  \nfor a specified region\\.",
    ),
    IntrinsicFunction(
        function="ImportValue",
        documentation="The intrinsic function `Fn::ImportValue` returns the value of [an output exported](outputs-section-structure.md)  \nby another stack\\. You typically use this function to  \n[create cross\\-stack references](walkthrough-crossstackref.md)\\. In the following example template snippets,  \nStack A exports VPC security group values and Stack B imports them\\.",
    ),
    IntrinsicFunction(
        function="Join",
        documentation="The intrinsic function `Fn::Join` appends a set of values into a single value,  \nseparated by the specified delimiter\\. If a delimiter is the empty string, the  \nset of values are concatenated with no delimiter\\.",
    ),
    IntrinsicFunction(
        function="Select",
        documentation="The intrinsic function `Fn::Select` returns a single object from a list of  \nobjects by index\\.",
    ),
    IntrinsicFunction(
        function="Split",
        documentation="To split a string into a list of string values so that you can select an element  \nfrom the resulting string list, use the `Fn::Split` intrinsic function\\. Specify  \nthe location of splits with a delimiter, such as `,` \\(a comma\\)\\. After you  \nsplit a string, use the [`Fn::Select`](intrinsic-function-reference-select.md) function to pick a specific element\\.",
    ),
    IntrinsicFunction(
        function="Sub",
        documentation="The intrinsic function `Fn::Sub` substitutes variables in an input string with  \nvalues that you specify\\. In your templates, you can use this function to  \nconstruct commands or outputs that include values that aren't available until you  \ncreate or update a stack\\.",
    ),
    IntrinsicFunction(
        function="Transform",
        documentation="The intrinsic function `Fn::Transform` specifies a macro to perform custom  \nprocessing on part of a stack template\\. Macros enable you to perform custom  \nprocessing on templates, from simple actions like find\\-and\\-replace operations  \nto extensive transformations of entire templates\\. For more information, see  \n[Using AWS CloudFormation macros to perform custom processing on templates](template-macros.md)\\.",
    ),
    IntrinsicFunction(
        function="And",
        documentation="Returns `true` if all the specified conditions evaluate to true, or returns  \n`false` if any one of the conditions evaluates to false\\. `Fn::And` acts as  \nan AND operator\\. The minimum number of conditions that you can include is 2,  \nand the maximum is 10\\.",
    ),
    IntrinsicFunction(
        function="Equals",
        documentation="Compares if two values are equal\\. Returns `true` if the two values are  \nequal or `false` if they aren't\\.",
    ),
    IntrinsicFunction(
        function="If",
        documentation="For the `Fn::If` function, you only need to specify the condition name\\. The  \nfollowing snippet shows how to use `Fn::If` to conditionally specify a resource  \nproperty\\. If the `CreateLargeSize` condition is true, CloudFormation sets the  \nvolume size to `100`\\. If the condition is false, CloudFormation sets the volume  \nsize to `10`\\.",
    ),
    IntrinsicFunction(
        function="Not",
        documentation="Returns `true` for a condition that evaluates to `false` or returns `false`  \nfor a condition that evaluates to `true`\\. `Fn::Not` acts as a NOT operator\\.",
    ),
    IntrinsicFunction(
        function="Or",
        documentation="Returns `true` if any one of the specified conditions evaluate to true, or  \nreturns `false` if all the conditions evaluates to false\\. `Fn::Or` acts as an OR  \noperator\\. The minimum number of conditions that you can include is 2, and the maximum  \nis 10\\.",
    ),
    IntrinsicFunction(
        function="Ref",
        full_name_prefix="",
        documentation="The intrinsic function `Ref` returns the value of the specified *parameter* or  \n*resource*\\. When the [`AWS::LanguageExtensions` transform](transform-aws-languageextensions.md) transform is used, you  \ncan use intrinsic functions as a parameter to [`Ref`](#intrinsic-function-reference-ref) and [`Fn::GetAtt`](intrinsic-function-reference-getatt.md)\\. \n + When you specify a parameter's logical name, it returns the value of the parameter\\.\n + When you specify a resource's logical name, it returns a value that you can  \ntypically use to refer to that resource, such as a [physical ID](resources-section-structure.md)\\. \n + When you specify an intrinsic function, it returns the output of that function\\.",
    ),
]



"""
Completion logic.
"""

from lsprotocol.types import CompletionItem, CompletionList, Position
from pygls.workspace import Document

from ..aws_data import AWSContext, AWSPropertyName, Tree
from ..cursor import text_edit, word_before_after_position
from ..decode.extractors import (
    AllowedValuesExtractor,
    ResourceExtractor,
    ResourcePropertyExtractor,
)
from .allowed_values import allowed_values_completions
from .attributes import attribute_completions
from .functions import intrinsic_function_completions
from .ref import ref_completions
from .resources import resource_completions
from .static import static_completions

TRIGGER_CHARACTERS = [
    ".",
    "Type: ",
    "!Ref ",
    "Ref: ",
    "!GetAtt ",
    "GetAtt: ",
    "!",
    '"Type": "',
    '"Ref": "',
    '"GetAtt": "',
    '"',
]


def completions_for(
    template_data: Tree,
    aws_context: AWSContext,
    document: Document,
    position: Position,
    allowed_values_extractor: AllowedValuesExtractor,
    use_sam: bool,
) -> CompletionList:
    """Return a list of completion items for the user's position in document."""
    line, char = position.line, position.character
    resource_lookup = ResourceExtractor().extract(template_data)
    res_span = resource_lookup.at(line, char)
    if res_span:
        return resource_completions(res_span.value, aws_context, document, position)

    allowed_values_completions_result = allowed_values_completions(
        template_data, aws_context, document, position, allowed_values_extractor
    )
    if allowed_values_completions_result:
        return allowed_values_completions_result

    prop_lookup = ResourcePropertyExtractor().extract(template_data)
    prop_span = prop_lookup.at(line, char)
    if prop_span:
        return property_completions(prop_span.value, aws_context, document, position)

    ref_completions_result = ref_completions(
        template_data, document, position, aws_context
    )
    if ref_completions_result:
        return ref_completions_result

    att_completions_result = attribute_completions(
        template_data, aws_context, document, position
    )
    if att_completions_result:
        return att_completions_result

    static_completions_result = static_completions(
        template_data, aws_context, document, position, use_sam
    )
    if static_completions_result:
        return static_completions_result

    return intrinsic_function_completions(document, position)


def property_completions(
    name: AWSPropertyName,
    aws_context: AWSContext,
    document: Document,
    position: Position,
) -> CompletionList:
    if name.parent in aws_context:
        before, after = word_before_after_position(document, position)
        suffix = (
            ""
            if (document.filename is not None and document.filename.endswith("json"))
            else ": "
        )
        return CompletionList(
            is_incomplete=False,
            items=[
                CompletionItem(
                    label=s.short_form(),
                    documentation=aws_context.description(s),
                    text_edit=text_edit(
                        position, before, after, s.short_form() + suffix
                    ),
                )
                for s in aws_context.same_level(name)
            ],
        )
    return CompletionList(is_incomplete=False, items=[])



"""
Intrinsic function completions, see:
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html
For more information.
"""
from typing import Callable, List

from lsprotocol.types import CompletionItem, CompletionList, Position
from pygls.workspace import Document

from ..cursor import text_edit, word_before_after_position
from ..intrinsic_functions import INTRINSIC_FUNCTIONS, IntrinsicFunction


def intrinsic_function_completions(
    document: Document,
    position: Position,
    intrinsic_functions: List[IntrinsicFunction] = INTRINSIC_FUNCTIONS,
) -> CompletionList:
    before, after = word_before_after_position(document, position)
    word = before + after
    filter_fn: Callable[[IntrinsicFunction], bool]
    label_fn: Callable[[IntrinsicFunction], str]
    if word.startswith(("Fn", "fn")):
        filter_fn = lambda f: f.full_name_prefix == "Fn::"
        label_fn = lambda f: f.full_name()
    elif word.startswith("!"):
        filter_fn = lambda f: f.short_form_prefix == "!"
        label_fn = lambda f: f.short_form()
    else:
        return CompletionList(is_incomplete=False, items=[])

    items = [
        CompletionItem(
            label=label_fn(f),
            text_edit=text_edit(position, before, after, label_fn(f) + " "),
            documentation=f.documentation,
        )
        for f in intrinsic_functions
        if filter_fn(f)
    ]
    return CompletionList(is_incomplete=False, items=items)



"""Command-line interface."""

import logging
from pathlib import Path
from typing import Optional

import click
from click import Context

from .context import load_cfn_context, load_sam_context
from .server import server

logger = logging.getLogger(__name__)


@click.group(invoke_without_command=True)
@click.version_option()
@click.option("-v", "--verbose", count=True, help="Print more output.")
@click.pass_context
def cli(ctx: Context, verbose: int) -> None:
    """Start a cfn-lsp-extra server."""
    # This fn is called regardless, so we have to check if a subcommand should be run
    level = [logging.ERROR, logging.INFO, logging.DEBUG][min(verbose, 2)]
    logging.basicConfig(level=level, force=True)
    if ctx.invoked_subcommand is None:
        cfn_aws_context = load_cfn_context()
        sam_aws_context = load_sam_context(cfn_aws_context)
        logger.info("Starting cfn-lsp-extra server")
        server(cfn_aws_context, sam_aws_context).start_io()


@cli.command()
@click.option("-s", "--specification-file", type=click.Path(exists=True))
@click.option("-d", "--documentation-directory", type=click.Path(exists=True))
@click.option("-o", "--output-path", type=click.Path(exists=False))
def update_specification(
    specification_file: Optional[str] = None,
    documentation_directory: Optional[str] = None,
    output_path: Optional[str] = None,
) -> None:
    """Update the specification used by cfn-lsp-extra."""
    logging.basicConfig(level=logging.INFO, force=True)
    try:
        from .scrape.specification import run
    except ImportError as e:
        raise Exception(
            "Please Install cfn-lsp-extra[parse] to run 'update-specification'"
        ) from e
    else:
        run(
            Path(specification_file) if specification_file else None,
            Path(documentation_directory).absolute()
            if documentation_directory
            else None,
            Path(output_path) if output_path else None,
        )


@cli.command()
@click.option("-s", "--specification-file", type=click.Path(exists=True))
@click.option("-d", "--documentation-directory", type=click.Path(exists=True))
@click.option("-o", "--output-path", type=click.Path(exists=False))
def update_sam_specification(
    specification_file: Optional[str] = None,
    documentation_directory: Optional[str] = None,
    output_path: Optional[str] = None,
) -> None:
    logging.basicConfig(level=logging.INFO, force=True)
    try:
        from .scrape.sam_specification import run
    except ImportError as e:
        raise Exception(
            "Please Install cfn-lsp-extra[parse] to run 'update-specification'"
        ) from e
    else:
        run(
            Path(specification_file) if specification_file else None,
            Path(documentation_directory).absolute()
            if documentation_directory
            else None,
            Path(output_path) if output_path else None,
        )


def main() -> None:
    cli(prog_name="cfn-lsp-extra", auto_envvar_prefix="CFN_LSP_EXTRA")


if __name__ == "__main__":
    main()



"""Cfn Lsp Extra."""


# credit https://stackoverflow.com/questions/16891340/remove-a-prefix-from-a-string
# TODO remove when support for Python 3.8 is dropped
def remove_prefix(text: str, prefix: str) -> str:
    """Remove prefix from text if necessary."""
    if text.startswith(prefix):
        return text[len(prefix) :]
    return text


# TODO remove when support for Python 3.8 is dropped
def remove_suffix(text: str, suffix: str) -> str:
    """Remove suffix from text if necessary."""
    if text.endswith(suffix):
        return text[: -len(suffix)]
    return text



import json

import yaml
from lsprotocol.types import Position

from ..aws_data import Tree
from .json_decoding import CfnJSONDecoder  # type: ignore[attr-defined]
from .yaml_decoding import SafePositionLoader

DEBUG_CHAR = "."


class CfnDecodingError(Exception):
    pass


def decode(source: str, filename: str) -> Tree:
    """Deserialise the cloudformation template source into a dictionary.

    Parameters
    ----------
    source : str
        template content, expected to be either json or yaml.
    filename: str
        name of the template file, determined whether source is taken to
        be json or yaml.

    Returns
    -------
    Tree
        A recursive dictionary containing items and positional information
        from source.

    Raises
    ------
    CfnDecodingError
        If json or yaml parsing fails."""
    try:
        if filename.endswith("json"):
            data = json.loads(source, cls=CfnJSONDecoder)
        else:
            data = yaml.load(source, Loader=SafePositionLoader)  # noqa
    except (json.JSONDecodeError, yaml.YAMLError) as e:
        raise CfnDecodingError(f"Error decoding {filename}") from e
    return data


def decode_unfinished(source: str, filename: str, position: Position) -> Tree:
    """Deserialise the cloudformation template source into a dictionary.

    If decoding fails, attempt to 'fix' source by making edits to the
    current line.  If position is on an empty line, the character at
    position will be set to an '.' (to aid completions).

    Parameters
    ----------
    source : str
        template content, expected to be either json or yaml.
    filename: str
        name of the template file, determined whether source is taken to
        be json or yaml.
    position: Position
        The position of the user's cursor in the document

    Returns
    -------
    Tree
        A recursive dictionary containing items and positional information
        from source.

    Raises
    ------
    CfnDecodingError
        If json or yaml parsing fails even after edits."""
    source_lst = source.splitlines()
    line, char = position.line, position.character
    if not filename.endswith("json"):
        source_lst[line] = yaml_line_enricher(source_lst[line], char)
        source = "\n".join(source_lst)
    try:
        return decode(source, filename)
    except CfnDecodingError:
        if filename.endswith("json"):
            source_lst[line] = source_lst[line].rstrip(":, ") + ': "",'
        else:
            source_lst[line] = source_lst[line].rstrip() + ":"
        source = "\n".join(source_lst)
        return decode(source, filename)


def yaml_line_enricher(line: str, char: int) -> str:
    stripped = line.strip()
    if not stripped:  # Empty line for property
        new_line = line[:char] + DEBUG_CHAR
    elif stripped.endswith((":", "!Ref", "!GetAtt")):  # For resource or ref
        new_line = line + DEBUG_CHAR
    elif stripped.startswith("-"):  # First element in a list
        new_line = line + DEBUG_CHAR + ": "
    else:
        new_line = line
    return new_line



"""
Definitions for !Refs.
"""

import logging
from typing import Optional

from lsprotocol.types import Location, Position
from pygls.workspace import Document

from ..aws_data import AWSContext, Tree
from .attributes import attribute_definition
from .ref import ref_definition

logger = logging.getLogger(__name__)


def definition(
    template_data: Tree,
    document: Document,
    position: Position,
    aws_context: AWSContext,
) -> Optional[Location]:
    ref_result = ref_definition(template_data, document, position, aws_context)
    logger.error(ref_result)
    if ref_result:
        return ref_result

    attribute_result = attribute_definition(
        template_data, document, position, aws_context
    )
    logger.error(attribute_result)
    return attribute_result



"""
Intrinsic function hovers, see:
https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/intrinsic-function-reference.html
For more information.
"""
from typing import List, Optional

from lsprotocol.types import Hover, MarkupContent, MarkupKind, Position, Range
from pygls.workspace import Document

from ..cursor import word_before_after_position
from ..intrinsic_functions import INTRINSIC_FUNCTIONS, IntrinsicFunction


def intrinsic_function_hover(
    document: Document,
    position: Position,
    intrinsic_functions: List[IntrinsicFunction] = INTRINSIC_FUNCTIONS,
) -> Optional[Hover]:
    before, after = word_before_after_position(document, position)
    word = before + after

    if word.startswith(("Fn", "fn")):
        fn = next(
            (
                f
                for f in INTRINSIC_FUNCTIONS
                if f.full_name() == word or word.startswith(f.full_name() + ":")
            ),
            None,
        )
    elif word.startswith("!"):
        fn = next((f for f in INTRINSIC_FUNCTIONS if f.short_form() == word), None)
    else:
        return None

    if fn:
        char_start = position.character - len(before)
        return Hover(
            range=Range(
                start=Position(line=position.line, character=char_start),
                end=Position(line=position.line, character=char_start + len(word)),
            ),
            contents=MarkupContent(kind=MarkupKind.Markdown, value=fn.documentation),
        )
    return None



"""
Static completions, e.g. at the top level we can only complete 'AWSTemplateFormatVersion', 'Resources',
etc.
"""
from typing import Optional

from lsprotocol.types import CompletionItem, CompletionList, Position
from pygls.workspace import Document

from ..aws_data import AWSContext, Tree
from ..cursor import text_edit, word_before_after_position
from ..decode.extractors import StaticExtractor, StaticPath

# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/template-anatomy.html
TOP_LEVEL_ATTRIBUTES = [
    "AWSTemplateFormatVersion",
    "Description",
    "Metadata",
    "Parameters",
    "Rules",
    "Mappings",
    "Conditions",
    "Transform",
    "Resources",
    "Outputs"
]

TOP_LEVEL_PATH = StaticPath.root(StaticPath.MatchAny)

# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/parameters-section-structure.html
PARAMETER_ATTRIBUTES = [
    "Type",
    "Default",
    "AllowedValues",
    "Description"
]

PARAMETER_PATH = StaticPath.root("Parameters") / StaticPath.MatchAny / StaticPath.MatchAny

RESOURCE_ATTRIBUTES = [
    "Type",
    "Properties",
    # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-product-attribute-reference.html
    "CreationPolicy",
    "DeletionPolicy",
    "DependsOn",
    "Metadata",
    "UpdatePolicy",
    "UpdateReplacePolicy"
]

RESOURCE_PATH = StaticPath.root("Resources") / StaticPath.MatchAny / StaticPath.MatchAny

# https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/outputs-section-structure.html
OUTPUT_ATTRIBUTES = [
    "Description",
    "Value",
    "Export"
]

OUTPUT_PATH = StaticPath.root("Outputs") / StaticPath.MatchAny / StaticPath.MatchAny

TOP_LEVEL_SAM_ATTRIBUTES = [
    "Transform",
    "Globals",
    "Description",
    "Metadata",
    "Parameters",
    "Mappings",
    "Conditions",
    "Resources",
    "Outputs"
]

TOP_LEVEL_SAM_PATH = StaticPath.root(StaticPath.MatchAny)

COMMON_STATIC_LOOKUP = {
    PARAMETER_PATH: PARAMETER_ATTRIBUTES,
    RESOURCE_PATH: RESOURCE_ATTRIBUTES,
    OUTPUT_PATH: OUTPUT_ATTRIBUTES,
    TOP_LEVEL_SAM_PATH: TOP_LEVEL_SAM_ATTRIBUTES,
}
CFN_STATIC_LOOKUP = {**COMMON_STATIC_LOOKUP, TOP_LEVEL_PATH: TOP_LEVEL_ATTRIBUTES}
SAM_STATIC_LOOKUP = {**COMMON_STATIC_LOOKUP, TOP_LEVEL_SAM_PATH: TOP_LEVEL_SAM_ATTRIBUTES}
CFN_EXTRACTOR = StaticExtractor(paths=set(CFN_STATIC_LOOKUP.keys()))
SAM_EXTRACTOR = StaticExtractor(paths=set(SAM_STATIC_LOOKUP.keys()))


def static_completions(
    template_data: Tree,
    aws_context: AWSContext,
    document: Document,
    position: Position,
    use_sam: bool
) -> Optional[CompletionList]:
    lookup, extractor = (SAM_STATIC_LOOKUP, SAM_EXTRACTOR) if use_sam else (CFN_STATIC_LOOKUP, CFN_EXTRACTOR)
    position_lookup = extractor.extract(template_data)
    span = position_lookup.at(position.line, position.character)
    if span and span.value in lookup:
        before, after = word_before_after_position(document, position)
        suffix = "" if (document.filename is not None and document.filename.endswith("json")) else ": "
        return CompletionList(
            is_incomplete=False,
            items=[
                CompletionItem(
                    label=s,
                    text_edit=text_edit(position, before, after, s + suffix)
                )
                for s in lookup[span.value]
            ],
        )
    return None



"""
Utilities for parsing yaml document strings.
"""
from typing import Any, List, Tuple

from cfnlint.decode.cfn_yaml import FN_PREFIX, UNCONVERTED_SUFFIXES
from cfnlint.decode.node import dict_node, list_node

try:
    from yaml.cyaml import CSafeLoader as SafeLoader
except ImportError:
    from yaml.loader import SafeLoader  # type: ignore[assignment]

from yaml.nodes import MappingNode, Node, ScalarNode, SequenceNode

POSITION_PREFIX = "__position__"
VALUES_POSITION_PREFIX = "__value_positions__"


# Copied from an earlier version of cfnlint for compat
def multi_constructor(loader: SafeLoader, tag_suffix: str, node: Node) -> Node:
    """Deal with !Ref style function format."""

    if tag_suffix not in UNCONVERTED_SUFFIXES:
        tag_suffix = f"{FN_PREFIX}{tag_suffix}"

    if tag_suffix == "Fn::GetAtt":
        constructor = construct_getatt
    elif isinstance(node, ScalarNode):
        constructor = loader.construct_scalar  # type: ignore[assignment]
    elif isinstance(node, SequenceNode):
        constructor = loader.construct_sequence  # type: ignore[assignment]
    elif isinstance(node, MappingNode):
        constructor = loader.construct_mapping  # type: ignore[assignment]
    else:
        raise Exception(f"Bad tag: !{tag_suffix}")

    return dict_node({tag_suffix: constructor(node)}, node.start_mark, node.end_mark)  # type: ignore[no-any-return]


def construct_getatt(node: Node) -> Node:
    """Reconstruct !GetAtt into a list."""

    if isinstance(node.value, str):
        return list_node(node.value.split(".", 1), node.start_mark, node.end_mark)  # type: ignore[no-any-return]
    if isinstance(node.value, list):
        return list_node([s.value for s in node.value], node.start_mark, node.end_mark)  # type: ignore[no-any-return]

    raise ValueError(f"Unexpected node type: {type(node.value)}")


class SafePositionLoader(SafeLoader):
    """A loader which saves positional information on elements.

    It takes inspiration from https://stackoverflow.com/questions/13319067/parsing-yaml-return-with-line-number#13319530.
    """  # noqa

    yaml_multi_constructors = {"!": multi_constructor}

    def _positional_key_value(self, node: Node) -> Tuple[str, List[int]]:
        return POSITION_PREFIX + node.value, [
            node.start_mark.line,
            node.start_mark.column,
        ]

    def _value_position(self, scalar_node: ScalarNode) -> Any:
        position_key, position_value = self._positional_key_value(scalar_node)
        return {position_key: position_value}

    def construct_mapping(self, node: MappingNode, deep: bool = False) -> Any:
        mapping = super(SafePositionLoader, self).construct_mapping(node, deep=deep)

        for key_node, value_node in node.value:
            position_key, position_value = self._positional_key_value(key_node)
            mapping[position_key] = position_value

            # Stick positional info on leaf nodes into their own mapping so they
            # aren't confused with others, first we check if value_node is a
            # !Ref
            obj = self.construct_object(value_node)  # type: ignore[no-untyped-call]
            if isinstance(value_node, ScalarNode) and hasattr(obj, "start_mark"):
                key = key_node.value
                if not isinstance(mapping[key], dict):
                    continue  # This is a duplicate key (ie bad template)
                to_add = []
                for _, sub_value in obj.items():
                    line, char = value_node.end_mark.line, value_node.end_mark.column
                    sub_value = (
                        ".".join(sub_value)
                        if isinstance(sub_value, list)
                        else sub_value
                    )
                    char -= len(sub_value)
                    to_add.append({POSITION_PREFIX + value_node.value: [line, char]})
                mapping[key][VALUES_POSITION_PREFIX] = to_add

            # node is a normal ScalarNode (not a !Ref)
            elif isinstance(value_node, ScalarNode):
                if VALUES_POSITION_PREFIX not in mapping:
                    mapping[VALUES_POSITION_PREFIX] = []
                value_position = self._value_position(value_node)
                mapping[VALUES_POSITION_PREFIX].append(value_position)
        return mapping

    def construct_sequence(self, node: SequenceNode, deep: bool = False) -> Any:
        mapping = super(SafePositionLoader, self).construct_sequence(node, deep)

        for idx, value_node in enumerate(node.value):
            obj = self.construct_object(value_node)  # type: ignore[no-untyped-call]
            if isinstance(value_node, ScalarNode) and hasattr(obj, "start_mark"):
                to_add = []
                for _, sub_value in obj.items():
                    line, char = value_node.end_mark.line, value_node.end_mark.column
                    char -= len(sub_value)
                    to_add.append({POSITION_PREFIX + value_node.value: [line, char]})
                mapping[idx][VALUES_POSITION_PREFIX] = to_add

        return mapping



"""
Definitions for !Refs.
"""

from typing import Optional

from lsprotocol.types import Location, Position, Range
from pygls.workspace import Document

from ..aws_data import AWSContext, Tree
from ..ref import resolve_ref


def ref_definition(
    template_data: Tree,
    document: Document,
    position: Position,
    aws_context: AWSContext,
) -> Optional[Location]:
    link = resolve_ref(position, template_data)
    if link:
        return Location(
            uri=document.uri,
            range=Range(
                start=Position(
                    line=link.source_span.line, character=link.source_span.char
                ),
                end=Position(
                    line=link.source_span.line,
                    character=link.source_span.char + link.source_span.span,
                ),
            ),
        )
    return None



from __future__ import annotations

import re
from textwrap import TextWrapper
from typing import List, Pattern


class MarkdownLink(str):
    def __new__(cls, url: str, description: str) -> MarkdownLink:
        obj = str.__new__(cls, f"[{description}]({url})")
        obj.url = url  # type: ignore[attr-defined]
        obj.description = description  # type: ignore[attr-defined]
        return obj

    def __len__(self) -> int:
        return len(self.description)  # type: ignore[attr-defined]


class MarkdownTextWrapper(TextWrapper):
    """A TextWrapper which handles markdown links."""

    LINK_REGEX: Pattern[str] = re.compile(r"(\[.*?\]\(\S+\))")
    LINK_PARTS_REGEX: Pattern[str] = re.compile(r"^\[(.*?)\]\((\S+)\)$")

    def _split(self, text: str) -> List[str]:
        split = re.split(self.LINK_REGEX, text)
        chunks: List[str] = []
        for item in split:
            match = re.match(self.LINK_PARTS_REGEX, item)
            if match:
                chunks.append(MarkdownLink(match.group(2), match.group(1)))
            else:
                chunks.extend(super()._split(item))
        return chunks


TEXT_WRAPPER = MarkdownTextWrapper(
    width=79,
    break_long_words=True,
    replace_whitespace=False,
)



"""
Hovers for !GetAtt
"""
import re
from typing import Optional

from lsprotocol.types import Hover, MarkupContent, MarkupKind, Position, Range
from pygls.workspace import Document

from ..aws_data import AWSContext, AWSLogicalId, AWSResourceName, Tree
from ..cursor import word_at_position_char_bounds
from ..decode.extractors import Extractor, GetAttExtractor, LogicalIdExtractor

GET_ATT_EXTRACTOR = GetAttExtractor()
GET_ATT_SRC_EXTRACTOR = LogicalIdExtractor()
RE_END_ATTRIBUTE = re.compile(r"^[A-Za-z_0-9\.]*")
RE_START_ATTRIBUTE = re.compile(r"[A-Za-z_0-9\.]*$")


def attribute_hover(
    template_data: Tree,
    aws_context: AWSContext,
    document: Document,
    position: Position,
    get_att_extractor: Extractor[str] = GET_ATT_EXTRACTOR,
    get_att_src_extractor: Extractor[AWSLogicalId] = GET_ATT_SRC_EXTRACTOR,
) -> Optional[Hover]:
    get_att_lookup = get_att_extractor.extract(template_data)
    get_att_span = get_att_lookup.at(position.line, position.character)
    if get_att_span:
        text = get_att_span.value
        _, c, span = next(ps[0] for e, ps in get_att_lookup.items() if e == text)
        res, _, att = text.partition(".")

        get_att_src_lookup = get_att_src_extractor.extract(template_data)
        for get_att_src, _ in get_att_src_lookup.items():
            if get_att_src.logical_name == res:
                type_ = get_att_src.type_
                if type_:
                    resource_name = AWSResourceName(value=type_)
                    return_values = aws_context.return_values(resource_name)
                    char_start, char_end = word_at_position_char_bounds(
                        document, position, RE_START_ATTRIBUTE, RE_END_ATTRIBUTE
                    )
                    line_at = position.line
                    if att in return_values:
                        return Hover(
                            range=Range(
                                start=Position(line=line_at, character=char_start),
                                end=Position(line=line_at, character=char_end),
                            ),
                            contents=MarkupContent(
                                kind=MarkupKind.Markdown, value=return_values[att]
                            ),
                        )
                    break
    return None



"""
Utilities for loading the aws cloudformation doc content.
"""
import json
import logging
from collections import ChainMap
from pathlib import Path

from importlib_resources import as_file, files
from platformdirs import PlatformDirs

from .aws_data import AWSContext, AWSSpecification, Tree

logger = logging.getLogger(__name__)
dirs = PlatformDirs("cfn-lsp-extra", "cfn-lsp-extra")
CFN_OVERRIDE_CTX_PATH = Path(dirs.user_config_dir) / "context.json"
SAM_OVERRIDE_CTX_PATH = Path(dirs.user_config_dir) / "sam_context.json"
custom_ctx_path = Path(dirs.user_config_dir) / "custom.json"


def with_custom(context_map: Tree, custom_path: Path = custom_ctx_path) -> AWSContext:
    """Overwrite part of context with custom content."""
    logger.info("Updating context using custom file %s", custom_path)
    source = files("cfn_lsp_extra.resources").joinpath("custom.json")
    with as_file(source) as path, open(path, "r") as f:
        context_map = ChainMap(json.load(f), context_map)
    if custom_path.exists():
        logger.info("Updating context using user custom file %s", custom_path)
        with custom_path.open("r") as f:
            context_map = ChainMap(json.load(f), context_map)
    return AWSContext(
        resource_map=context_map[AWSSpecification.RESOURCE_TYPES],
        property_map=context_map[AWSSpecification.PROPERTY_TYPES],
    )


def load_context(
    resource: str, override_path: Path = CFN_OVERRIDE_CTX_PATH
) -> AWSContext:
    """Load AWS context from a cache."""
    if override_path.exists():
        logger.info("Loading custom context from %s", override_path)
        with override_path.open("r") as f:
            d = json.load(f)
            return AWSContext(
                d[AWSSpecification.RESOURCE_TYPES], d[AWSSpecification.PROPERTY_TYPES]
            )
    else:
        logger.info("Loading context...")
        source = files("cfn_lsp_extra.resources").joinpath(resource)
        with as_file(source) as path, open(path, "r") as f:
            return with_custom(json.load(f))


def load_cfn_context() -> AWSContext:
    return load_context("context.json", CFN_OVERRIDE_CTX_PATH)


def load_sam_context(cfn_context: AWSContext) -> AWSContext:
    return load_context("sam_context.json", SAM_OVERRIDE_CTX_PATH)



"""
Completions for property values.
"""
from typing import Optional

from lsprotocol.types import CompletionItem, CompletionList, Position
from pygls.workspace import Document

from ..aws_data import AWSContext, Tree
from ..cursor import text_edit, word_before_after_position
from ..decode.extractors import AllowedValuesExtractor


def allowed_values_completions(
    template_data: Tree,
    aws_context: AWSContext,
    document: Document,
    position: Position,
    allowed_values_extractor: AllowedValuesExtractor,
) -> Optional[CompletionList]:
    lookup = allowed_values_extractor.extract(template_data)
    span = lookup.at(position.line, position.character)
    if span and span.value in aws_context:
        allowed_values = aws_context.allowed_values(span.value)
        if allowed_values:
            before, after = word_before_after_position(document, position)
            items = [
                CompletionItem(
                    label=value,
                    text_edit=text_edit(position, before, after, value),
                )
                for value in allowed_values
            ]
            return CompletionList(is_incomplete=False, items=items)
    return None



"""
LSP implementation
https://pygls.readthedocs.io/en/latest/
https://microsoft.github.io/language-server-protocol/specifications/specification-current/
"""

import logging
import os
import re
import sys
from typing import Optional, Union

from lsprotocol.types import (
    COMPLETION_ITEM_RESOLVE,
    INITIALIZED,
    TEXT_DOCUMENT_COMPLETION,
    TEXT_DOCUMENT_DEFINITION,
    TEXT_DOCUMENT_DID_CHANGE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
    TEXT_DOCUMENT_HOVER,
    WORKSPACE_DID_CHANGE_CONFIGURATION,
    CompletionItem,
    CompletionList,
    CompletionOptions,
    CompletionParams,
    DefinitionParams,
    DidChangeConfigurationParams,
    DidChangeTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    Hover,
    HoverParams,
    InitializedParams,
    Location,
)
from pygls.server import LanguageServer
from pygls.workspace import Document

from .aws_data import AWSContext, AWSPropertyName, AWSResourceName
from .cfnlint_integration import CFNLINT_VERSION, diagnostics, load_cfnlint_config
from .completions import TRIGGER_CHARACTERS, completions_for
from .completions.resources import resolve_resource_completion_item
from .config.user_configuration import (
    DiagnosticPublishingMethod,
    UserConfiguration,
    configuration_params,
    from_did_change_config,
    from_get_configuration_response,
)
from .decode import CfnDecodingError, decode, decode_unfinished
from .decode.extractors import (
    AllowedValuesExtractor,
    CompositeExtractor,
    ResourceExtractor,
    ResourcePropertyExtractor,
)
from .definitions import definition
from .hovers import hover

logger = logging.getLogger(__name__)


def server(cfn_aws_context: AWSContext, sam_aws_context: AWSContext) -> LanguageServer:
    server = LanguageServer("cfn-lsp-extra", "")  # TODO get real version here
    allowed_values_extractor = AllowedValuesExtractor()
    extractor = CompositeExtractor[Union[AWSResourceName, AWSPropertyName]](
        ResourcePropertyExtractor(), ResourceExtractor()
    )
    config = UserConfiguration()
    logger.info("PYTHONPATH: %s", os.environ.get("PYTHONPATH"))
    logger.info("sys.path: %s", sys.path)
    logger.info("cfnlint version: %s", CFNLINT_VERSION)
    logger.info("Test loading cfnlint configuration...")
    load_cfnlint_config(log_exceptions=True)

    @server.thread()
    @server.feature(INITIALIZED)
    def intialiazed(ls: LanguageServer, params: InitializedParams) -> None:
        """Client initialized notification."""
        workspace_capabilities = ls.client_capabilities.workspace
        if workspace_capabilities and workspace_capabilities.configuration:
            logger.info("Obtaining user config")
            try:
                config_response = ls.get_configuration(
                    params=configuration_params()
                ).result(timeout=1)
            except TimeoutError:
                logger.error("Timeout retrieving config")
            else:
                nonlocal config
                config = from_get_configuration_response(config_response)
                logger.info("Obtained user config: %s", config)

    @server.thread()
    @server.feature(TEXT_DOCUMENT_DID_OPEN)
    def did_open(ls: LanguageServer, params: DidOpenTextDocumentParams) -> None:
        """Text document did open notification."""
        uri = params.text_document.uri
        text_doc = ls.workspace.get_document(uri)  # type: ignore[no-untyped-call]
        logger.debug("Is template SAM: %s", is_document_sam(text_doc))
        file_path = text_doc.path
        ls.publish_diagnostics(text_doc.uri, diagnostics(text_doc.source, file_path))

    @server.thread()
    @server.feature(TEXT_DOCUMENT_DID_CHANGE)
    def did_change(ls: LanguageServer, params: DidChangeTextDocumentParams) -> None:
        """Text document did change notification."""
        uri = params.text_document.uri
        text_doc = ls.workspace.get_document(uri)  # type: ignore[no-untyped-call]
        file_path = text_doc.path
        if (
            config.diagnostic_publishing_method
            == DiagnosticPublishingMethod.ON_DID_CHANGE
        ):
            # Publishing diagnostics removes old ones
            ls.publish_diagnostics(
                text_doc.uri, diagnostics(text_doc.source, file_path)
            )

    @server.thread()
    @server.feature(TEXT_DOCUMENT_DID_SAVE)
    def did_save(ls: LanguageServer, params: DidSaveTextDocumentParams) -> None:
        """Text document did save notification."""
        uri = params.text_document.uri
        text_doc = ls.workspace.get_document(uri)  # type: ignore[no-untyped-call]
        file_path = text_doc.path
        if (
            config.diagnostic_publishing_method
            == DiagnosticPublishingMethod.ON_DID_SAVE
        ):
            ls.publish_diagnostics(
                text_doc.uri, diagnostics(text_doc.source, file_path)
            )

    @server.feature(
        TEXT_DOCUMENT_COMPLETION,
        CompletionOptions(trigger_characters=TRIGGER_CHARACTERS, resolve_provider=True),
    )
    def completions(
        ls: LanguageServer, params: CompletionParams
    ) -> Optional[CompletionList]:
        """Returns completion items."""
        uri = params.text_document.uri
        document = server.workspace.get_document(uri)  # type: ignore[no-untyped-call]
        use_sam = is_document_sam(document)
        aws_context = sam_aws_context if use_sam else cfn_aws_context
        try:
            template_data = decode_unfinished(
                document.source, document.filename, params.position
            )
        except CfnDecodingError as e:
            logger.debug("Failed to decode document: %s", e)
            return None
        return completions_for(
            template_data,
            aws_context,
            document,
            params.position,
            allowed_values_extractor,
            use_sam,
        )

    @server.feature(COMPLETION_ITEM_RESOLVE)
    def completion_item_resolve(
        ls: LanguageServer, completion_item: CompletionItem
    ) -> CompletionItem:
        """Resolves a completion item."""
        if re.match(r"^.+::.+::.+$", completion_item.label):
            return resolve_resource_completion_item(completion_item, cfn_aws_context)
        return completion_item  # Not a resource

    @server.feature(TEXT_DOCUMENT_HOVER)
    def did_hover(ls: LanguageServer, params: HoverParams) -> Optional[Hover]:
        """Text document did hover notification."""
        uri = params.text_document.uri
        document = server.workspace.get_document(uri)  # type: ignore[no-untyped-call]
        aws_context = sam_aws_context if is_document_sam(document) else cfn_aws_context
        try:
            template_data = decode(document.source, document.filename)
        except CfnDecodingError as e:
            logger.debug("Failed to decode document: %s", e)
            return None
        position_lookup = extractor.extract(template_data)
        return hover(
            template_data, params.position, aws_context, document, position_lookup
        )

    @server.feature(TEXT_DOCUMENT_DEFINITION)
    def goto_definition(
        ls: LanguageServer, params: DefinitionParams
    ) -> Optional[Location]:
        document = server.workspace.get_document(params.text_document.uri)  # type: ignore[no-untyped-call]
        aws_context = sam_aws_context if is_document_sam(document) else cfn_aws_context
        try:
            template_data = decode(document.source, document.filename)
        except CfnDecodingError as e:
            logger.debug("Failed to decode document: %s", e)
            return None
        return definition(template_data, document, params.position, aws_context)

    @server.feature(WORKSPACE_DID_CHANGE_CONFIGURATION)
    def did_change_configuration(
        ls: LanguageServer, params: DidChangeConfigurationParams
    ) -> None:
        """Workspace did change configuration notification."""
        logger.info("Received user configuration: %s", params)
        nonlocal config
        config = from_did_change_config(params)

    return server


def is_document_sam(document: Document) -> bool:
    for line in document.lines:
        line_stripped = line.strip()
        if not line_stripped.startswith("#") and not line_stripped.startswith("{"):
            return (
                line_stripped == "Transform: AWS::Serverless-2016-10-31"
                or line_stripped.replace(" ", "").startswith(
                    '"Transform":"AWS::Serverless-2016-10-31"'
                )
            )
    return False



"""
Classes for dealing with aws properties.

For reference as of 21/05/2022 there are 186 resource prefixes. and a total
of 899 resources.  The most properties for a resource is 52 for
'AWS::RDS::DBInstance' and the average is 6.193548387096774.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional, Union

from attrs import frozen

from .scrape.markdown_textwrapper import TEXT_WRAPPER

# A Tree type representing a recursive nested structure such as yaml or json
# https://github.com/python/mypy/issues/731
# Tree = Dict[str, Union[str, "Tree"]]
Tree = Any


class AWSRoot(Enum):
    """Represents the root of the heirarchy.

    This enum's root value is this type's sole value."""

    root = 0

    def parent(self) -> AWSRoot:
        return self

    def __truediv__(self, resource: str) -> AWSResourceName:
        return AWSResourceName(value=resource)


@frozen
class AWSResourceName:
    value: str
    parent: AWSRoot = AWSRoot.root

    def __str__(self) -> str:
        return self.value

    def __truediv__(self, other: str) -> AWSPropertyName:
        return AWSPropertyName(parent=self, property_=other)

    def split(self) -> List[str]:
        return [self.value]

    def short_form(self) -> str:
        return self.value


@frozen
class AWSPropertyName:
    """Property of an AWS resource or AWS resource property."""

    parent: Union[AWSPropertyName, AWSResourceName]
    property_: str

    def __str__(self) -> str:
        return f"{self.parent}/{self.property_}"

    def __truediv__(self, other: str) -> AWSPropertyName:
        return AWSPropertyName(parent=self, property_=other)

    def __contains__(self, other: str) -> bool:
        if other == self.property_:
            return True
        if isinstance(self.parent, AWSResourceName):
            return other == self.parent.value
        return other in self.parent

    def split(self) -> List[str]:
        if isinstance(self.parent, AWSResourceName):
            return [str(self.parent), self.property_]
        return self.parent.split() + [self.property_]

    def short_form(self) -> str:
        return self.property_


AWSName = Union[AWSResourceName, AWSPropertyName]


class AWSSpecification:
    """Fields from the AWS Cloudformation resource and property specification"""

    RESOURCE_TYPES = "ResourceTypes"
    PROPERTY_TYPES = "PropertyTypes"
    ATTRIBUTES = "Attributes"
    PROPERTIES = "Properties"
    DOCUMENTATION = "Documentation"
    REQUIRED = "Required"
    TYPE = "Type"
    PRIMITIVE_TYPE = "Type"
    ITEM_TYPE = "ItemType"

    # Syntethic Fields
    MARKDOWN_DOCUMENTATION = "MarkdownDocumentation"
    ALLOWED_VALUES = "AllowedValues"
    REF_RETURN_VALUE = "RefReturnValue"


class AWSContext:
    """A handle on AWS resource data for the lsp server."""

    def __init__(self, resource_map: Tree, property_map: Tree):
        self.resource_map = resource_map
        """For resources that have properties within a property (also known as subproperties), a list of subproperty specifications"""
        self.property_map = property_map
        self.property_map_lc = {k.lower(): k for k in property_map}

    def __enrich_name(self, name: AWSName) -> AWSName:
        resource, *props = name.split()
        name = AWSResourceName(resource)
        for prop in props:
            name = name / prop
            tree = self.__raw_getitem(name)
            if AWSSpecification.ITEM_TYPE in tree:
                name = name / tree[AWSSpecification.ITEM_TYPE]
        return name

    def __raw_getitem(self, name: AWSName) -> Tree:
        try:
            resource, *props = name.split()
            tree = self.resource_map[resource]
            if props:  # Is NOT a resource
                idx = -1
                for idx, prop in list(enumerate(props))[::-1]:  # noqa: B007
                    property_key = f"{resource}.{prop}"
                    if property_key in self.property_map:
                        tree = self.property_map[property_key]
                        break
                    if property_key.lower() in self.property_map_lc:
                        tree = self.property_map[
                            self.property_map_lc[property_key.lower()]
                        ]
                        break
                else:
                    idx = -1

                for nested_prop in props[idx + 1 :]:
                    tree = tree[AWSSpecification.PROPERTIES][nested_prop]

            return tree
        except KeyError as e:
            raise KeyError(f"'{name}' is not a recognised resource or property") from e

    def __getitem__(self, name: AWSName) -> Tree:
        try:
            return self.__raw_getitem(self.__enrich_name(name))
        except KeyError:
            # Sometimes ItemType is something like 'String' or 'Tag', which results in an invalid name
            return self.__raw_getitem(name)

    def __contains__(self, name: AWSName) -> bool:
        try:
            self[name]
        except KeyError:
            return False
        else:
            return True

    def description(self, name: AWSName) -> str:
        """Get the description of obj."""
        # Be a bit forgiving here a la SAM specification
        return self[name].get(AWSSpecification.MARKDOWN_DOCUMENTATION, "")  # type: ignore[no-any-return]

    def return_values(self, resource: AWSResourceName) -> Dict[str, str]:
        dcts = self[resource].get(AWSSpecification.ATTRIBUTES, {})
        return {
            k: v.get(AWSSpecification.MARKDOWN_DOCUMENTATION, "")
            for k, v in dcts.items()
        }

    def ref_return_value(self, resource: AWSResourceName) -> str:
        return self[resource].get(AWSSpecification.REF_RETURN_VALUE, "unknown")  # type: ignore[no-any-return]

    def allowed_values(self, property_: AWSPropertyName) -> List[str]:
        return self[property_].get(AWSSpecification.ALLOWED_VALUES, [])  # type: ignore[no-any-return]

    def same_level(self, obj: AWSName) -> List[AWSName]:
        """Return names at the same (property/resource) level as obj."""
        if isinstance(obj, AWSResourceName):
            return [AWSResourceName(value=r) for r in self.resource_map]
        if isinstance(obj, AWSPropertyName):
            try:
                return [
                    obj.parent / p
                    for p in self[obj.parent][AWSSpecification.PROPERTIES]
                ]
            except KeyError:
                return []
        raise ValueError(f"obj has to be of type AWSName, but was '{type(obj)}'")


@frozen
class AWSRefName:
    value: str


@frozen
class AWSRefSource(ABC):
    """An object which can be linked to using !Ref, e.g. a parameter."""

    logical_name: str

    @abstractmethod
    def as_documentation(self, aws_context: AWSContext) -> str:
        ...


@frozen
class AWSParameter(AWSRefSource):
    logical_name: str
    type_: str
    description: Optional[str] = None
    default: Optional[str] = None

    def as_documentation(self, aws_context: AWSContext) -> str:
        description_str = "\n" + self.description if self.description else ""
        # We can't triple quote here since flake8 will flag the "  " as trailing whitespace...
        return (
            f"### Parameter: `{self.logical_name}`{description_str}  \n"
            f"*Type*: `{self.type_}`  \n"
            f"*Default*: {self.default}"
        )


@frozen
class AWSLogicalId(AWSRefSource):
    logical_name: str
    type_: Optional[str]

    def as_documentation(self, aws_context: AWSContext) -> str:
        ref_return_value = ""
        if self.type_:
            res_name = AWSResourceName(value=self.type_)
            if res_name in aws_context:
                ref_return_value = (
                    f"\n*Return Value*: {aws_context.ref_return_value(res_name)}  "
                )
        type_str = "`" + self.type_ + "`" if self.type_ else "not given"
        full_str = (
            f"### Resource: `{self.logical_name}`  \n"
            f"*Type*: {type_str}{ref_return_value}"
        )
        content = ""
        for line in full_str.splitlines():
            content += "\n".join(TEXT_WRAPPER.wrap(line)) + "\n"
        return content



"""
Utilities from tracking total failures/successes over the lifespan of some process.

This would be so nice with a writer monad.
"""
from typing import Generic, Tuple, TypeVar

from attrs import frozen

E = TypeVar("E")
E2 = TypeVar("E2")


@frozen
class WithSuccessFailureCount(Generic[E]):
    """Wraps an object along with a success/failure count.

    Attributes
    ----------
    value : E
        The wrapped value
    successes : int
        The number of successes.
    failures : int
        The number of failures."""

    value: E
    successes: int
    failures: int

    @classmethod
    def zero(cls, e: E) -> "WithSuccessFailureCount[E]":
        return WithSuccessFailureCount(e, 0, 0)

    @classmethod
    def success(cls, e: E) -> "WithSuccessFailureCount[E]":
        return WithSuccessFailureCount(e, 1, 0)

    @classmethod
    def failure(cls, e: E) -> "WithSuccessFailureCount[E]":
        return WithSuccessFailureCount(e, 0, 1)

    def add_counts(
        self, other: "WithSuccessFailureCount[E2]"
    ) -> "WithSuccessFailureCount[E]":
        return WithSuccessFailureCount(
            self.value,
            self.successes + other.successes,
            self.failures + other.failures,
        )

    def to_tuple(self) -> Tuple[E, int, int]:
        return (self.value, self.successes, self.failures)



"""
Completions for !GetAtt
"""
from typing import Optional

from lsprotocol.types import CompletionItem, CompletionList, Position
from pygls.workspace import Document

from ..aws_data import AWSContext, AWSLogicalId, AWSResourceName, Tree
from ..cursor import text_edit, word_before_after_position
from ..decode.extractors import Extractor, GetAttExtractor, LogicalIdExtractor

GET_ATT_EXTRACTOR = GetAttExtractor()
GET_ATT_SRC_EXTRACTOR = LogicalIdExtractor()


def attribute_completions(
    template_data: Tree,
    aws_context: AWSContext,
    document: Document,
    position: Position,
    get_att_extractor: Extractor[str] = GET_ATT_EXTRACTOR,
    get_att_src_extractor: Extractor[AWSLogicalId] = GET_ATT_SRC_EXTRACTOR,
) -> Optional[CompletionList]:
    get_att_lookup = get_att_extractor.extract(template_data)
    get_att_span = get_att_lookup.at(position.line, position.character)
    if get_att_span:
        text = get_att_span.value
        _, col, _ = next(ps[0] for e, ps in get_att_lookup.items() if e == text)
        res, _, att = text.partition(".")

        before, after = word_before_after_position(document, position)
        if col <= position.character <= col + len(res):
            get_att_src_lookup = get_att_src_extractor.extract(template_data)
            items = [
                CompletionItem(
                    label=get_att_src.logical_name,
                    documentation=get_att_src.as_documentation(aws_context),
                    text_edit=text_edit(
                        position, before, after, get_att_src.logical_name
                    ),
                )
                for get_att_src, _ in get_att_src_lookup.items()
            ]
            return CompletionList(is_incomplete=False, items=items)

        get_att_src_lookup = get_att_src_extractor.extract(template_data)
        for get_att_src, _ in get_att_src_lookup.items():
            if get_att_src.logical_name == res:
                type_ = get_att_src.type_
                if type_:
                    resource_name = AWSResourceName(value=type_)
                    if type_ and resource_name in aws_context:
                        items = [
                            CompletionItem(label=return_val, documentation=desc)
                            for return_val, desc in aws_context.return_values(
                                resource_name
                            ).items()
                        ]
                        return CompletionList(is_incomplete=False, items=items)
    return None



from typing import Optional

from lsprotocol.types import Position

from .aws_data import AWSRefName, AWSRefSource, Tree
from .decode.extractors import (
    CompositeExtractor,
    Extractor,
    KeyExtractor,
    LogicalIdExtractor,
    ParameterExtractor,
)
from .decode.position import PositionLink, Spanning

REF_EXTRACTOR = KeyExtractor[AWSRefName]("Ref", lambda s: AWSRefName(value=s))
REF_SRC_EXTRACTOR = CompositeExtractor[AWSRefSource](
    LogicalIdExtractor(), ParameterExtractor()
)


def resolve_ref(
    position: Position,
    template_data: Tree,
    ref_extractor: Extractor[AWSRefName] = REF_EXTRACTOR,
    ref_src_extractor: Extractor[AWSRefSource] = REF_SRC_EXTRACTOR,
) -> Optional[PositionLink[AWSRefSource, AWSRefName]]:
    """Attempt to resolve the source and documentation of a ref at position.

    Parameters
    ----------
    position : Position
        The position to look at in the template
    template_data : Tree
        The decoded template
    extractor: Extractor[AWSRefSource]
        The extractor to use to extract the source of refs

    Returns
    -------
    Optional[PositionLink[AWSRefSource, AWSRefName]]
        A spanning object containing the ref source if it was found, else None
    """
    ref_lookup = ref_extractor.extract(template_data)
    ref_span = ref_lookup.at(position.line, position.character)
    if ref_span:
        ref_src_lookup = ref_src_extractor.extract(template_data)
        for ref_src, position_list in ref_src_lookup.items():
            if ref_src.logical_name == ref_span.value.value and position_list:
                line, char, _ = position_list[0]
                src_span = Spanning[AWSRefSource](
                    value=ref_src, line=line, char=char, span=len(ref_src.logical_name)
                )
                return PositionLink[AWSRefSource, AWSRefName](
                    source_span=src_span, target_span=ref_span
                )
    return None



"""
Hovers for !Refs.
"""
from typing import Optional, Union

from lsprotocol.types import Hover, MarkupContent, MarkupKind, Position, Range

from ..aws_data import AWSContext, AWSPropertyName, AWSResourceName, Tree
from ..decode.position import PositionLookup
from ..ref import resolve_ref


def ref_hover(
    template_data: Tree,
    position: Position,
    aws_context: AWSContext,
    position_lookup: PositionLookup[Union[AWSResourceName, AWSPropertyName]],
) -> Optional[Hover]:
    # Attempt to resolve it as a Ref
    link = resolve_ref(position, template_data)
    if link:
        documentation = link.source_span.value.as_documentation(aws_context)
        char, length = link.target_span.char, link.target_span.span
        line_at = position.line
        return Hover(
            range=Range(
                start=Position(line=line_at, character=char),
                end=Position(line=line_at, character=char + length),
            ),
            contents=MarkupContent(kind=MarkupKind.Markdown, value=documentation),
        )
    return None



"""
Utilities for handling user configuration.

See also:
https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#workspace_configuration
https://pygls.readthedocs.io/en/latest/pages/advanced_usage.html#configuration
"""
from enum import Enum, auto, unique
from typing import Any, List

from attrs import define
from lsprotocol.types import (
    ConfigurationItem,
    ConfigurationParams,
    DidChangeConfigurationParams,
    WorkspaceConfigurationParams,
)

SECTION = "cfn"


@unique
class DiagnosticPublishingMethod(Enum):
    ON_DID_SAVE = auto()
    ON_DID_CHANGE = auto()


DIAGNOSTIC_PUBLISHING_METHOD_KEY = "diagnosticPublishingMethod"
DIAGNOSTIC_PUBLISHING_METHOD_DEFAULT = DiagnosticPublishingMethod.ON_DID_CHANGE


@define
class UserConfiguration:
    diagnostic_publishing_method: DiagnosticPublishingMethod = (
        DIAGNOSTIC_PUBLISHING_METHOD_DEFAULT
    )


# TODO in 3.11+ https://docs.python.org/3/library/enum.html#enum.StrEnum is much nicer for this
def diagnostic_publishing_method_from_string(string: str) -> DiagnosticPublishingMethod:
    if string.upper() == "ON_DID_SAVE":
        return DiagnosticPublishingMethod.ON_DID_SAVE
    return DIAGNOSTIC_PUBLISHING_METHOD_DEFAULT


def workspace_configuration_params(uri: str) -> WorkspaceConfigurationParams:
    return WorkspaceConfigurationParams(
        items=[
            ConfigurationItem(scope_uri=uri, section=DIAGNOSTIC_PUBLISHING_METHOD_KEY)
        ]
    )


def configuration_params() -> ConfigurationParams:
    items = [ConfigurationItem(section=f"{SECTION}.{DIAGNOSTIC_PUBLISHING_METHOD_KEY}")]
    return ConfigurationParams(items=items)


def from_get_configuration_response(config: List[Any]) -> UserConfiguration:
    """Obtain config from the response of a workspace/configuration response.

    https://microsoft.github.io/language-server-protocol/specifications/lsp/3.17/specification/#workspace_configuration
    """
    user_config = UserConfiguration()
    if config and isinstance(config, list) and len(config) > 0:
        diagnostic_publishing_method, *_ = config
        if diagnostic_publishing_method and isinstance(
            diagnostic_publishing_method, str
        ):
            user_config.diagnostic_publishing_method = (
                diagnostic_publishing_method_from_string(diagnostic_publishing_method)
            )
    return user_config


def from_did_change_config(config: DidChangeConfigurationParams) -> UserConfiguration:
    """Obtain config from a DidChangeConfigurationParams object.

    Example object:
    DidChangeConfigurationParams(settings={'cfn': {'diagnosticPublishingMethod': 'ON_DID_CHANGE'}})
    """
    user_config = UserConfiguration()
    if config.settings:
        relevant_settings = config.settings.get(SECTION, {})
        user_config.diagnostic_publishing_method = (
            diagnostic_publishing_method_from_string(
                relevant_settings.get(
                    DIAGNOSTIC_PUBLISHING_METHOD_KEY,
                    str(DIAGNOSTIC_PUBLISHING_METHOD_DEFAULT),
                )
            )
        )
    return user_config



"""
https://raw.githubusercontent.com/awslabs/goformation/master/schema/sam.schema.json
"""
import json  # noqa: I001
import logging
import os
import tempfile
from pathlib import Path
from typing import Optional

import requests

from ..aws_data import AWSSpecification, Tree
from .specification import (
    PARSE_SUCCESS_RATIO_KEY,
    documentation,
    file_content,
    run_command,
)
from .with_success_failure_count import WithSuccessFailureCount

logger = logging.getLogger(__name__)

PROP_KEY = "properties"
REF_KEY = "$ref"
REF_PREFIX = "#/definitions/"
REQUIRED = "required"
TYPE_KEY = "type"
PRIMITIVE_TYPE = "PrimitiveType"
PROPERTY_SKIP_LIST = (
    REQUIRED,
    PRIMITIVE_TYPE,
    "additionalProperties",
    PROP_KEY,
    TYPE_KEY,
)
ANY_OF = "anyOf"
SERVERLESS_PREFIX = "AWS::Serverless"
DEFAULT_SPEC_URL = "https://raw.githubusercontent.com/awslabs/goformation/master/schema/sam.schema.json"
DOC_URL = "docs.aws.amazon.com/serverless-application-model/latest/developerguide/"


def to_aws_context(
    sam_dct: Tree, base_directory: Path, base_url: str
) -> WithSuccessFailureCount[Tree]:
    d_: WithSuccessFailureCount[Tree] = WithSuccessFailureCount.zero(
        {AWSSpecification.RESOURCE_TYPES: {}, AWSSpecification.PROPERTY_TYPES: {}}
    )
    base = sam_dct["definitions"]
    for k, v in base.items():
        if "." not in k and SERVERLESS_PREFIX in k:
            res_with_counts = normalise_resource(k, v, base_directory, base_url)
            d_.value[AWSSpecification.RESOURCE_TYPES][k] = res_with_counts.value
            d_ = d_.add_counts(res_with_counts)
        elif SERVERLESS_PREFIX in k:
            res_with_counts = normalise_property(k, v, base_directory, base_url)
            d_.value[AWSSpecification.PROPERTY_TYPES][k] = res_with_counts.value
            d_ = d_.add_counts(res_with_counts)
    return d_


def normalise_resource(
    name: str, d: Tree, base_directory: Path, base_url: str
) -> WithSuccessFailureCount[Tree]:
    if not isinstance(d, dict):
        return WithSuccessFailureCount.zero(d)

    d_: WithSuccessFailureCount[Tree] = WithSuccessFailureCount.zero({})
    props = d[PROP_KEY][AWSSpecification.PROPERTIES][PROP_KEY]
    d_.value[AWSSpecification.PROPERTIES] = props
    link = f"{base_url}sam-resource-{name.split('::')[-1].lower()}"
    bs = file_content(base_directory, link)
    doc_with_counts = documentation(bs, link, name)
    d_.value[AWSSpecification.MARKDOWN_DOCUMENTATION] = doc_with_counts.value
    d_ = d_.add_counts(doc_with_counts)

    required = d[PROP_KEY][AWSSpecification.PROPERTIES].get(REQUIRED, [])
    for property_name in list(props.keys()):
        sub_props = d_.value[AWSSpecification.PROPERTIES][property_name]
        sub_props[AWSSpecification.REQUIRED] = property_name in required
        sub_doc_with_counts = documentation(bs, link, property_name)
        sub_props[AWSSpecification.MARKDOWN_DOCUMENTATION] = sub_doc_with_counts.value
        d_ = d_.add_counts(sub_doc_with_counts)
        normalise_types(sub_props)
    return d_


def normalise_property(
    name: str, d: Tree, base_directory: Path, base_url: str
) -> WithSuccessFailureCount[Tree]:
    if not isinstance(d, dict):
        return WithSuccessFailureCount.zero(d)

    parent, _, prop = name.partition(".")
    resource = parent.split("::")[-1]

    d_: WithSuccessFailureCount[Tree] = WithSuccessFailureCount.zero({})
    d_.value[AWSSpecification.PROPERTIES] = d[PROP_KEY]

    parent_link = f"{base_url}sam-resource-{parent.split('::')[-1].lower()}"
    parent_bs = file_content(base_directory, parent_link)
    a_element = parent_bs.find(
        "a",
        string=lambda x: x and x.lower() == prop.lower(),
        href=lambda h: not h.startswith("#") and "#" in h,
    )
    if a_element:
        link_val = a_element if isinstance(a_element, str) else a_element["href"]
        link = link_val[0] if isinstance(link_val, list) else link_val
        logger.info("Found replacement URL for %s: %s", name, link)
    else:
        link = f"{base_url}sam-property-{resource.lower()}-{prop.lower()}"
    bs = file_content(base_directory, link)
    doc_with_counts = documentation(bs, link, name)
    d_.value[AWSSpecification.MARKDOWN_DOCUMENTATION] = doc_with_counts.value
    d_ = d_.add_counts(doc_with_counts)

    for property_name, sub_props in d_.value[AWSSpecification.PROPERTIES].items():
        # property_name = 'Type', v =  {'type': 'string'}
        sub_doc_with_counts = documentation(bs, link, property_name)
        sub_props[AWSSpecification.MARKDOWN_DOCUMENTATION] = sub_doc_with_counts.value
        d_ = d_.add_counts(sub_doc_with_counts)
        normalise_types(sub_props)
    return d_


def normalise_types(d: Tree) -> Tree:
    if REF_KEY in d:
        ref = d.pop(REF_KEY)
        d[AWSSpecification.TYPE] = ref.split(".")[-1]
    elif ANY_OF in d:  # We just take the first
        types = d.pop(ANY_OF)
        find = next((t[REF_KEY].split(".")[-1] for t in types if REF_KEY in t), None)
        if find:
            d[AWSSpecification.TYPE] = find
        else:
            d[AWSSpecification.PRIMITIVE_TYPE] = types[0]
    elif TYPE_KEY in d and d[TYPE_KEY] == "array":
        d.pop(TYPE_KEY)
        d[AWSSpecification.TYPE] = "List"
        item_type = d.pop("items")
        if isinstance(item_type, dict) and REF_KEY in item_type:
            d[AWSSpecification.ITEM_TYPE] = item_type[REF_KEY].split(".")[-1]
        else:
            d[AWSSpecification.ITEM_TYPE] = item_type[TYPE_KEY]
    elif TYPE_KEY in d:
        d[AWSSpecification.PRIMITIVE_TYPE] = d.pop(TYPE_KEY)
    return d


def run(
    spec_file: Optional[Path],
    documentation_directory: Optional[Path],
    out_path: Optional[Path],
) -> None:
    out_file = Path("new-aws-sam-context.json").absolute()
    if spec_file:
        spec_json = json.loads(spec_file.read_bytes())
    else:
        logger.info("Downloading spec from %s", DEFAULT_SPEC_URL)
        spec_json = json.loads(requests.get(DEFAULT_SPEC_URL).text)
    with tempfile.TemporaryDirectory() as tmp_directory:
        os.chdir(tmp_directory)
        doc_dir = documentation_directory or Path(DOC_URL)
        base_url = f"https://{DOC_URL}"
        if not documentation_directory:
            run_command(f"wget --no-parent -r {base_url}")
        logger.info("Using base URL: %s", base_url)
        aws_context, md_succ, md_fails = to_aws_context(
            sam_dct=spec_json, base_directory=doc_dir, base_url=base_url
        ).to_tuple()
        logger.info("Failed getting markdown: %d/%d", md_fails, md_fails + md_succ)
        aws_context[PARSE_SUCCESS_RATIO_KEY] = round(md_succ / (md_fails + md_succ), 4)
        with open(out_file, "w") as sam_spec_out:
            json.dump(aws_context, sam_spec_out, indent=2)
    logger.info("Wrote context to %s", out_file)



"""

"""
import logging
from typing import Optional, Union

from lsprotocol.types import Hover, MarkupContent, MarkupKind, Position, Range
from pygls.workspace import Document

from ..aws_data import AWSContext, AWSPropertyName, AWSResourceName, Tree
from ..decode.position import PositionLookup
from .attributes import attribute_hover
from .functions import intrinsic_function_hover
from .refs import ref_hover

logger = logging.getLogger(__name__)


def hover(
    template_data: Tree,
    position: Position,
    aws_context: AWSContext,
    document: Document,
    position_lookup: PositionLookup[Union[AWSResourceName, AWSPropertyName]],
) -> Optional[Hover]:
    line_at, char_at = position.line, position.character
    span = position_lookup.at(line_at, char_at)

    if span:
        try:
            documentation = aws_context.description(span.value)
        except KeyError:  # no description for value, e.g. incomplete
            pass
        else:
            char, length = span.char, span.span
            return Hover(
                range=Range(
                    start=Position(line=line_at, character=char),
                    end=Position(line=line_at, character=char + length),
                ),
                contents=MarkupContent(kind=MarkupKind.Markdown, value=documentation),
            )

    for_ref = ref_hover(template_data, position, aws_context, position_lookup)
    if for_ref:
        return for_ref

    for_attribute = attribute_hover(template_data, aws_context, document, position)
    if for_attribute:
        return for_attribute

    for_intrinsic_function = intrinsic_function_hover(document, position)
    if for_intrinsic_function:
        return for_intrinsic_function

    return None



"""
LSP features leveraging cfnlint.

https://github.com/aws-cloudformation/cfn-lint/blob/main/docs/getting_started/integration.md
"""
import logging
from typing import Dict, List

import cfnlint
from cfnlint.api import lint
from cfnlint.config import ConfigFileArgs
from lsprotocol.types import Diagnostic, DiagnosticSeverity, Position, Range

logger = logging.getLogger(__name__)

CFNLINT_VERSION = cfnlint.version.__version__


def diagnostics(yaml_content: str, file_path: str) -> List[Diagnostic]:
    """Return diagnostics for the template file at file_path."""

    try:
        config = load_cfnlint_config(log_exceptions=False)
    except Exception:
        config = {}

    matches = lint(yaml_content, config=config)

    severity_mapping: Dict[str, DiagnosticSeverity] = {
        "informational": DiagnosticSeverity.Information,
        "error": DiagnosticSeverity.Error,
        "warning": DiagnosticSeverity.Warning,
    }
    return [
        Diagnostic(
            range=Range(
                start=Position(line=m.linenumber - 1, character=m.columnnumber - 1),
                end=Position(line=m.linenumberend - 1, character=m.columnnumberend - 1),
            ),
            message=m.message,
            source="cfn-lsp-extra",
            severity=severity_mapping.get(
                m.rule.severity.lower(), DiagnosticSeverity.Error
            ),
        )
        for m in filter(match_filter, matches)
    ]


def load_cfnlint_config(log_exceptions: bool) -> Dict[str, str]:
    try:
        # https://github.com/aws-cloudformation/cfn-lint/blob/main/src/cfnlint/config.py
        config = ConfigFileArgs()
        config.load()
        return config.file_args  # type: ignore[no-any-return]
    except Exception as e:
        if log_exceptions:
            logger.error("Error reading cfnlint configuration %s", e)
        return {}


def match_filter(rule_match: cfnlint.rules.Match) -> bool:
    # TODO we should probably log this
    return not rule_match.message.lower().startswith(
        "unknown exception while processing rule"
    )



"""
Logic for handling resource completions.

There are 899+ resources in total, which empircally is ok to send to a
client as long as we don't send send the documentation along with the
labels and snippets.
"""
from lsprotocol.types import (
    CompletionItem,
    CompletionList,
    InsertTextFormat,
    MarkupContent,
    MarkupKind,
    Position,
)
from pygls.workspace import Document

from ..aws_data import AWSContext, AWSResourceName, AWSSpecification
from ..cursor import text_edit, word_before_after_position


def resource_completions(
    name: AWSResourceName,
    aws_context: AWSContext,
    document: Document,
    position: Position,
) -> CompletionList:
    """Return a list of all resources, without documentation."""
    use_snippet = (
        not document.filename or not document.filename.endswith("json")
    ) and (
        position.line == len(document.lines) - 1
        or not document.lines[position.line + 1].strip()
    )
    before, after = word_before_after_position(document, position)
    items = []
    for r in map(lambda r: r.short_form(), aws_context.same_level(name)):
        if use_snippet:
            insert = r + "\n" + resource_snippet(AWSResourceName(value=r), aws_context)
        else:
            insert = r
        items.append(
            CompletionItem(
                label=r,
                text_edit=text_edit(position, before, after, insert),
                insert_text_format=InsertTextFormat.Snippet if use_snippet else None,
            )
        )
    return CompletionList(is_incomplete=False, items=items)


def resolve_resource_completion_item(
    completion_item: CompletionItem, aws_context: AWSContext
) -> CompletionItem:
    """Enrich a completion_item with documentation."""
    resource_name = AWSResourceName(value=completion_item.label)
    completion_item.documentation = MarkupContent(
        kind=MarkupKind.Markdown, value=aws_context.description(resource_name)
    )
    return completion_item


def resource_snippet(name: AWSResourceName, aws_context: AWSContext) -> str:
    """Return a snippet appropriate for the resource name using aws_context."""
    props = "Properties:\n"
    required_props = (
        p
        for p, v in aws_context[name][AWSSpecification.PROPERTIES].items()
        if v[AWSSpecification.REQUIRED]
    )
    for idx, prop in enumerate(required_props):
        props += f"\t{prop}: ${idx + 1}\n"
    # $0 defines the final tab stop
    return props + "\t$0"
